import time

import numpy as np
from joblib import Parallel, delayed
from numba import njit
from PySide6 import QtCore

# from PySide6.QtCore import *
from scipy.ndimage import uniform_filter1d
from tqdm.autonotebook import tqdm

from piscat.InputOutput.cpu_configurations import CPUConfigurations
from piscat.Preproccessing import FPNc


@njit(parallel=True)
def numba_divide(ndarray_1, ndarray_2):
    divide_result = np.divide(ndarray_1, ndarray_2)
    return divide_result


@njit(parallel=True)
def numba_sum(video, axis):
    sum_result = np.sum(video, axis=axis)
    return sum_result


@njit(parallel=True)
def numba_diff(ndarray_1, ndarray_2):
    diff_result = ndarray_1 - ndarray_2
    return diff_result


class WorkerSignals(QtCore.QObject):
    updateProgress_DRA = QtCore.Signal(int)
    result_final = QtCore.Signal(object)
    finished_DRA = QtCore.Signal()
    DRA_complete_signal = QtCore.Signal(bool)
    updateProgress_FPNC = QtCore.Signal(int)


class DifferentialRollingAverage(QtCore.QRunnable):
    def __init__(
        self,
        video=None,
        batchSize=500,
        flag_GUI=False,
        object_update_progressBar=None,
        mode_FPN="mFPN",
        FPN_flag_GUI=False,
        gui_select_correction_axis=1,
    ):
        """Differential Rolling Average (DRA).

        Parameters
        ----------
        video: NDArray
            The video is 3D-numpy (number of frames, width, height).

        batchSize: int
            The number of frames in each batch.

        mode_FPN: {‘cpFPN’, ‘mFPN’, ‘wFPN’, 'fFPN'}, optional
            Flag that defines method of FPNc.

                    * `mFPN`: Median fixed pattern noise correction
                    * `cpFPN`: Median fixed pattern noise correction
                    * `wFPN`: Wavelet FPNc
                    * `fFPN`: FFT2D_Wavelet FPNc

        optional_1: GUI
            These flags are used when GUI calls this method.

            * `flag_GUI`: bool
                This flag is defined as True when GUI calls this method.

            * `FPN_flag_GUI`: bool
                This flag is defined as True when GUI calls this method while
                we want activate FPNc.

            * `gui_select_correction_axis`: int (0/1), 'Both'
                This parameter is used only when FPN_flag_GUI is True,
                otherwise it will be ignored.

            * `object_update_progressBar`: object
                Object that updates the progress bar in GUI.

        """
        super(DifferentialRollingAverage, self).__init__()

        self.cpu = CPUConfigurations()

        self.flag_GUI = flag_GUI
        self.mode_FPN = mode_FPN
        self.FPN_flag_run = FPN_flag_GUI
        self.DRA_flag_run = True
        self.select_correction_axis_run = gui_select_correction_axis

        self.size_A_diff = (video.shape[0] - 2 * batchSize, video.shape[1], video.shape[2])

        self.video = video.astype(np.float64)
        self.batchSize = batchSize

        self.moving_avg = np.empty_like(self.video)
        self.output_diff = np.empty(self.size_A_diff)

        self.output_batch_1 = np.empty(self.size_A_diff)
        self.fft_output_len = self.video.shape[0] - self.batchSize

        self.FPNc_video = None
        self.flag_thread_FPNc = True
        self.object_update_progressBar = object_update_progressBar
        self.threadpool = QtCore.QThreadPool()
        self.signals = WorkerSignals()

    @QtCore.Slot()
    def check_DRA_finish(self, flag):
        self.DRA_flag_run = flag

    @QtCore.Slot()
    def run(self, *args, **kwargs):
        self.signals.DRA_complete_signal.connect(self.check_DRA_finish)
        video_DRA = self.differential_rolling(
            FPN_flag=self.FPN_flag_run,
            select_correction_axis=self.select_correction_axis_run,
            FFT_flag=False,
            inter_flag_parallel_active=True,
        )

        self.signals.result_final.emit(video_DRA)

    @QtCore.Slot()
    def thread_FPNc_complete(self):
        self.object_update_progressBar.setRange(self.p_max)
        self.object_update_progressBar.setLabel("")
        print("THREAD FPNc COMPLETE!")

    @QtCore.Slot()
    def result_tread_FPNc(self, result):
        self.FPNc_video = result
        self.flag_thread_FPNc = False
        print("FPNc video update!")

    @QtCore.Slot()
    def startProgressBarFPNc(self, instance, **kwargs):
        self.thread_FPNc = instance(**kwargs)
        self.thread_FPNc.signals.updateProgress_FPNC.connect(
            self.object_update_progressBar.setProgress
        )
        self.thread_FPNc.signals.finished_FPNc.connect(self.thread_FPNc_complete)
        self.thread_FPNc.signals.result_FPNc.connect(self.result_tread_FPNc)

        self.threadpool.start(self.thread_FPNc)

    def differential_rolling(
        self,
        FPN_flag=False,
        select_correction_axis=1,
        FFT_flag=False,
        inter_flag_parallel_active=True,
        max_iterations=10,
        FFT_widith=1,
    ):
        """To use DRA, you'll need to call the "differential rolling" process.

        Parameters
        ----------
        FPN_flag: bool
            This flag activates the fixed pattern noise correction function in
            case define as true.

        select_correction_axis: int (0/1), 'Both'
            This parameter is used only when FPN_flag is True, otherwise it
            will be ignored.

            * 0: FPN will be applied row-wise.
            * 1: FPN will be applied column-wise.
            * 'Both': FPN will be applied on two axis.

        FFT_flag: bool
            In case it is True, DRA will be performed in parallel to improve
            the time performance.

        inter_flag_parallel_active: bool
            This flag actives/inactives parallel computation of wFPNc.

        max_iterations: int
            This parameter is used when fFPT is selected that defines the total
            number of filtering iterations.

        FFT_widith: int
            This parameter is used when fFPT is selected that defines the
            frequency mask's width.

        Returns
        -------
        output: NDArray
            Returns DRA video.

        gainMap1D_: NDArray
            Returns projection on each frame based on the correction axis

        """
        if FPN_flag and self.mode_FPN == "cpFPN":
            print("\n--- start DRA + cpFPN_axis: " + str(select_correction_axis) + "---")
        elif FPN_flag and self.mode_FPN == "wFPN":
            print("\n--- start DRA + wFPN_axis: " + str(select_correction_axis) + "---")
        elif FPN_flag and self.mode_FPN == "fFPN":
            print("\n--- start DRA + fFPN_axis: " + str(select_correction_axis) + "---")
        elif FPN_flag and self.mode_FPN == "mFPN":
            print("\n--- start DRA + mFPN_axis: " + str(select_correction_axis) + "---")
        else:
            print("\n--- start DRA ---")

        self._apply_moving_average(FFT_flag)

        if FPN_flag:
            if self.mode_FPN == "mFPN" or self.mode_FPN == "cpFPN":
                if self.flag_GUI is True:
                    self.DRA_flag_run = True

                    while self.DRA_flag_run:
                        QtCore.QCoreApplication.processEvents()

                    time.sleep(2)

                    video = self._FPNc_GUI_axis(
                        self.output_diff, select_correction_axis, self.mode_FPN
                    )

                    time.sleep(2)

                else:
                    out_diff = self.output_diff
                    video = self._FPNc_axis(
                        out_diff,
                        select_correction_axis,
                        self.mode_FPN,
                        inter_flag_parallel_active,
                    )
            else:
                video = self.output_diff
        else:
            video = self.output_diff

        output = numba_divide(video, self.output_batch_1)

        if FPN_flag:
            if self.mode_FPN == "wFPN" or self.mode_FPN == "fFPN":
                output = self._FPNc_axis(
                    output,
                    select_correction_axis,
                    self.mode_FPN,
                    inter_flag_parallel_active,
                    max_iterations,
                    FFT_widith,
                )

        if select_correction_axis == "Both":
            gainMap1D_1 = np.mean(output, axis=1)
            gainMap1D_2 = np.mean(output, axis=2)
            gainMap1D_ = [gainMap1D_1, gainMap1D_2]

        elif select_correction_axis is not None:
            gainMap1D_ = np.mean(output, axis=select_correction_axis + 1)

        else:
            gainMap1D_ = None

        return output, gainMap1D_

    def _apply_moving_average(self, FFT_flag):
        if FFT_flag:
            self.movingAvg_FFT_based()
        else:
            if self.flag_GUI:
                self.DRA_flag_run = True
                self.object_update_progressBar.setLabel("DRA")
            self.temporal_moving_average()

    def temporal_moving_average(self):
        batch_1 = np.sum(self.video[0 : self.batchSize, :, :], axis=0)
        batch_2 = np.sum(self.video[self.batchSize : 2 * self.batchSize, :, :], axis=0)

        batch_1_ = np.divide(batch_1, self.batchSize)
        batch_2_ = np.divide(batch_2, self.batchSize)

        self.output_diff[0, :, :] = batch_2_ - batch_1_
        self.output_batch_1[0, :, :] = batch_1_

        for i_ in tqdm(range(1, self.video.shape[0] - 2 * self.batchSize)):
            if self.flag_GUI is True:
                self.signals.updateProgress_DRA.emit(i_)
            batch_1 = (
                batch_1 - self.video[i_ - 1, :, :] + self.video[self.batchSize + i_ - 1, :, :]
            )
            batch_2 = (
                batch_2
                - self.video[self.batchSize + i_ - 1, :, :]
                + self.video[(2 * self.batchSize) + i_ - 1, :, :]
            )
            batch_1_ = np.divide(batch_1, self.batchSize)
            batch_2_ = np.divide(batch_2, self.batchSize)

            self.output_diff[i_, :, :] = batch_2_ - batch_1_
            self.output_batch_1[i_, :, :] = batch_1_

        if self.flag_GUI is True:
            self.signals.finished_DRA.emit()

    def numba_temporal_moving_average(self):
        batch_1 = numba_sum(self.video[0 : self.batchSize, :, :], axis=0)
        batch_2 = numba_sum(self.video[self.batchSize : 2 * self.batchSize, :, :], axis=0)

        batch_1_ = numba_divide(batch_1, self.batchSize)
        batch_2_ = numba_divide(batch_2, self.batchSize)

        self.output_diff[0, :, :] = numba_diff(batch_2_, batch_1_)
        self.output_batch_1[0, :, :] = batch_1_

        for i_ in tqdm(range(1, self.video.shape[0] - 2 * self.batchSize)):
            if self.flag_GUI is True:
                self.signals.updateProgress_DRA.emit(i_)
            batch_1 = (
                batch_1 - self.video[i_ - 1, :, :] + self.video[self.batchSize + i_ - 1, :, :]
            )
            batch_2 = (
                batch_2
                - self.video[self.batchSize + i_ - 1, :, :]
                + self.video[(2 * self.batchSize) + i_ - 1, :, :]
            )

            batch_1_ = numba_divide(batch_1, self.batchSize)
            batch_2_ = numba_divide(batch_2, self.batchSize)

            self.output_diff[0, :, :] = numba_diff(batch_2_, batch_1_)
            self.output_batch_1[0, :, :] = batch_1_

        if self.flag_GUI is True:
            self.signals.finished_DRA.emit()

    def movingAvg_FFT_based(self):
        Parallel(n_jobs=self.cpu.n_jobs, backend="threading")(
            delayed(self.uniform_filter1d_kernel)(r_, c_)
            for r_ in range(self.video.shape[1])
            for c_ in range(self.video.shape[2])
        )

        self.output_diff = (
            self.moving_avg[self.batchSize : self.fft_output_len, ...]
            - self.moving_avg[0 : self.fft_output_len - self.batchSize, ...]
        )

        self.output_batch_1 = self.moving_avg[0 : self.fft_output_len - self.batchSize, ...]

    def uniform_filter1d_kernel(self, r_, c_):
        self.moving_avg[:, r_, c_] = uniform_filter1d(
            self.video[:, r_, c_],
            size=self.batchSize,
            mode="constant",
            cval=0.0,
            origin=-(self.batchSize // 2),
            axis=0,
        )

    def _FPNc_axis(
        self,
        video,
        select_correction_axis,
        mode_FPN,
        inter_flag_parallel_active=True,
        max_iterations=10,
        FFT_widith=1,
    ):
        if mode_FPN == "cpFPN" or mode_FPN == "mFPN":
            if select_correction_axis == "Both":
                vid_FPNc_axis_1 = self._FPN(
                    video, 1, mode_FPN, inter_flag_parallel_active, max_iterations, FFT_widith
                )
                vid_FPNc = self._FPN(
                    vid_FPNc_axis_1,
                    0,
                    mode_FPN,
                    inter_flag_parallel_active,
                    max_iterations,
                    FFT_widith,
                )

            else:
                vid_FPNc = self._FPN(
                    video,
                    select_correction_axis,
                    mode_FPN,
                    inter_flag_parallel_active,
                    max_iterations,
                    FFT_widith,
                )
        else:
            vid_FPNc = self._FPN(
                video,
                select_correction_axis,
                mode_FPN,
                inter_flag_parallel_active,
                max_iterations,
                FFT_widith,
            )

        return vid_FPNc

    def _FPN(
        self,
        video,
        select_correction_axis,
        mode_FPN,
        inter_flag_parallel_active=True,
        max_iterations=10,
        FFT_widith=1,
    ):
        if mode_FPN == "mFPN":
            noise_c = FPNc.MedianProjectionFPNc(
                video=video, select_correction_axis=select_correction_axis, flag_GUI=self.flag_GUI
            )
            video_FPNc = noise_c.mFPNc(select_correction_axis)

        elif mode_FPN == "cpFPN":
            noise_c = FPNc.ColumnProjectionFPNc(
                video=video, select_correction_axis=select_correction_axis, flag_GUI=self.flag_GUI
            )
            video_FPNc = noise_c.cpFPNc(select_correction_axis)

        elif mode_FPN == "wFPN":
            if select_correction_axis == 0:
                wFPN = FPNc.FrequencyFPNc(
                    video, inter_flag_parallel_active=inter_flag_parallel_active
                )
                video_FPNc = wFPN.update_wFPN(direction="Horizontal")

            elif select_correction_axis == 1:
                wFPN = FPNc.FrequencyFPNc(
                    video, inter_flag_parallel_active=inter_flag_parallel_active
                )
                video_FPNc = wFPN.update_wFPN(direction="Vertical")

            elif select_correction_axis == "Both":
                wFPN = FPNc.FrequencyFPNc(
                    video, inter_flag_parallel_active=inter_flag_parallel_active
                )
                output = wFPN.update_wFPN(direction="Horizontal")

                wFPN = FPNc.FrequencyFPNc(
                    output, inter_flag_parallel_active=inter_flag_parallel_active
                )
                video_FPNc = wFPN.update_wFPN(direction="Vertical")

        elif mode_FPN == "fFPN":
            if select_correction_axis == 0:
                wf_FPN = FPNc.FrequencyFPNc(
                    video, inter_flag_parallel_active=inter_flag_parallel_active
                )
                video_FPNc = wf_FPN.update_fFPN(
                    direction="Horizontal", max_iterations=max_iterations, width=FFT_widith
                )

            elif select_correction_axis == 1:
                wf_FPN = FPNc.FrequencyFPNc(
                    video, inter_flag_parallel_active=inter_flag_parallel_active
                )
                video_FPNc = wf_FPN.update_fFPN(
                    direction="Vertical", max_iterations=max_iterations, width=FFT_widith
                )

            elif select_correction_axis == "Both":
                wf_FPN = FPNc.FrequencyFPNc(
                    video, inter_flag_parallel_active=inter_flag_parallel_active
                )
                output = wf_FPN.update_fFPN(
                    direction="Horizontal", max_iterations=max_iterations, width=FFT_widith
                )

                wf_FPN = FPNc.FrequencyFPNc(
                    output, inter_flag_parallel_active=inter_flag_parallel_active
                )
                video_FPNc = wf_FPN.update_fFPN(
                    direction="Vertical", max_iterations=max_iterations, width=FFT_widith
                )

        return video_FPNc

    def _FPNc_GUI_axis(self, video, select_correction_axis, mode_FPN):
        if select_correction_axis == "Both":
            self.flag_thread_FPNc = True
            vid_FPNc_axis_1 = self._FPN_GUI(video, 1, mode_FPN)

            while self.flag_thread_FPNc:
                QtCore.QCoreApplication.processEvents()

            time.sleep(2)

            self.flag_thread_FPNc = True
            vid_FPNc = self._FPN_GUI(vid_FPNc_axis_1, 0, mode_FPN)

            while self.flag_thread_FPNc:
                QtCore.QCoreApplication.processEvents()

            time.sleep(2)

        else:
            self.flag_thread_FPNc = True
            vid_FPNc = self._FPN_GUI(video, select_correction_axis, mode_FPN)

        return vid_FPNc

    def _FPN_GUI(self, video, select_correction_axis, mode_FPN):
        if mode_FPN == "mFPN":
            d_arg = {
                "video": video,
                "flag_GUI": True,
                "instance": FPNc.MedianProjectionFPNc,
                "select_correction_axis": select_correction_axis,
            }
        elif mode_FPN == "cpFPN":
            d_arg = {
                "video": video,
                "flag_GUI": True,
                "instance": FPNc.ColumnProjectionFPNc,
                "select_correction_axis": select_correction_axis,
            }

        self.p_max = video.shape[0] - (2 * self.batchSize) - 1
        self.object_update_progressBar.setRange(self.p_max)
        self.object_update_progressBar.setProgress(0)
        self.object_update_progressBar.setLabel("FPNc")

        self.flag_thread_FPNc = True
        self.startProgressBarFPNc(**d_arg)

        while self.flag_thread_FPNc:
            QtCore.QCoreApplication.processEvents()

        time.sleep(2)

        return self.FPNc_video



import numpy as np
from joblib import Parallel, delayed
from matplotlib import pyplot as plt

from piscat.BackgroundCorrection import DRA
from piscat.InputOutput.cpu_configurations import CPUConfigurations
from piscat.Visualization.print_colors import PrintColors


class NoiseFloor(CPUConfigurations, PrintColors):
    def __init__(
        self,
        video,
        list_range,
        FPN_flag=False,
        mode_FPN="mFPN",
        select_correction_axis=1,
        n_jobs=None,
        inter_flag_parallel_active=False,
        max_iterations=10,
        FFT_widith=1,
        mode="mode_temporal",
    ):
        """This class measures the noise floor for various batch sizes.

        Parameters
        ----------
        video: NDArray
            The video is 3D-numpy (number of frames, width, height).

        list_range: list
            list os all batch size that DRA should be calculated for them.

        FPN_flag: bool
            This flag activates the fixed pattern noise correction function in
            case define as true.

        mode_FPN: {‘cpFPN’, ‘mFPN’, ‘wFPN’, 'fFPN'}, optional
            Flag that defines method of FPNc.

            * `mFPN`: Median fixed pattern noise correction
            * `cpFPN`: Median fixed pattern noise correction
            * `wFPN`: Wavelet FPNc
            * `fFPN`: FFT2D_Wavelet FPNc

        select_correction_axis: int (0/1), 'Both'
            This parameter is used only when FPN_flag is True, otherwise it
            will be ignored.

            * `0`: FPN will be applied row-wise.
            * `1`: FPN will be applied column-wise.
            * `'Both'`: FPN will be applied on two axis.

        max_iterations: int
            This parameter is used when fFPT is selected that defines the total
            number of filtering iterations.

        FFT_widith: int
            This parameter is used when fFPT is selected that defines the
            frequency mask's width.

        """

        CPUConfigurations.__init__(self)
        PrintColors.__init__(self)

        self.video = video
        self.max_iterations = max_iterations
        self.FFT_widith = FFT_widith
        self.FPN_flag = FPN_flag
        self.select_correction_axis = select_correction_axis
        self.inter_flag_parallel_active = inter_flag_parallel_active
        self.mode_FPN = mode_FPN
        self.thr_shot_noise = None
        self.list_range = list_range
        if self.parallel_active and self.inter_flag_parallel_active:
            if n_jobs is not None:
                self.n_jobs = n_jobs
                print("\nThe number of usage CPU cores are {}!".format(n_jobs))

            self.mean = Parallel(n_jobs=self.n_jobs, backend=self.backend, verbose=0)(
                delayed(self.best_radius_kernel)(i_, flag_parallel=True, mode=mode)
                for i_ in range(len(self.list_range))
            )
        else:
            print(f"{self.WARNING}\nThe noise floor is running without parallel loop!{self.ENDC}")

            self.mean = []
            for i_ in range(len(self.list_range)):
                self.mean.append(self.best_radius_kernel(i_, flag_parallel=True, mode=mode))

    def best_radius_kernel(self, i_, flag_parallel, mode):
        DRA_ = DRA.DifferentialRollingAverage(
            video=self.video, batchSize=self.list_range[i_], mode_FPN=self.mode_FPN
        )
        video_DRA, _ = DRA_.differential_rolling(
            FPN_flag=self.FPN_flag,
            select_correction_axis=self.select_correction_axis,
            FFT_flag=False,
            inter_flag_parallel_active=flag_parallel,
            max_iterations=self.max_iterations,
            FFT_widith=self.FFT_widith,
        )

        if mode == "mode_temporal":
            noise_floor = np.mean(np.std(video_DRA, axis=0))
            # noise_floor = np.std(video_DRA, axis=0)
        elif mode == "mode_spatial":
            list_frame_std = []
            for f_ in range(video_DRA.shape[0]):
                frame_ = video_DRA[f_, ...]
                std_frame = np.std(np.ravel(frame_))
                list_frame_std.append(std_frame)

            noise_floor = np.mean(list_frame_std)

        return noise_floor

    def plot_result(self, flag_log=True):
        """
        The result of the noise floor is plotted when this function is called.

        Parameters
        ----------
        flag_log: bool
            The log-log plot style is enabled by this parameter.
        """
        list_batch = [1] + self.list_range[1:]
        shot_noise_ = np.divide(self.mean[0], np.sqrt(list_batch))
        if flag_log is False:
            fig, ax = plt.subplots()
            ax.plot(self.list_range, self.mean, "ro", label="Experimental result")
            ax.plot(self.list_range, shot_noise_, "b-", label="shot noise")

            ax.set_xlabel("Batch size")
            ax.set_ylabel("Noise floor")
            ax.legend()
            plt.show()

        if flag_log is True:
            fig, ax = plt.subplots()
            ax.loglog(self.list_range, self.mean, "ro", label="Experimental result")
            ax.loglog(self.list_range, shot_noise_, "b-", label="shot noise")

            ax.set_xlabel("Batch size")
            ax.set_ylabel("Noise floor")
            ax.legend()
            plt.show()



import os

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from PySide6 import QtCore, QtGui, QtWidgets
from PySide6.QtCore import QObject, QThreadPool, Signal, Slot

from piscat.GUI.InputOutput import save_GUI
from piscat.GUI.Visualization import slice_view
from piscat.GUI.Visualization.contrast_adjustment_GUI import Contrast_adjustment_GUI
from piscat.GUI.Visualization.play_setting import PlaySetting
from piscat.GUI.Visualization.temporal_annotation_setting import TemporalAnnotationSetting
from piscat.InputOutput import write_video
from piscat.Preproccessing.normalization import Normalization
from piscat.Visualization.display import Display, DisplayPSFs_subplotLocalizationDisplay
from piscat.Visualization.display_html import HTML_PSFs_subplotLocalizationDisplay

play_flag = True


class WorkerSignals(QObject):
    updateProgress = Signal(int)
    result = Signal(object)
    finished = Signal()
    stop_signal = Signal()


class ImageViewer(QtWidgets.QMainWindow, QtCore.QObject):
    progressChanged = QtCore.Signal(int)
    sliceNumber = QtCore.Signal(int)
    currentFrame = QtCore.Signal(int)

    def __init__(self, input_image_series, original_video, name, mask=False, list_psf=None):
        super(ImageViewer, self).__init__()

        self.threadpool = QThreadPool()

        self.image_series = input_image_series
        self.original_video = original_video
        self.df_PSFs = list_psf
        if self.image_series.ndim == 3:
            self.sizeX = self.image_series.shape[1]
            self.sizeY = self.image_series.shape[2]
        elif self.image_series.ndim == 2:
            self.sizeX = self.image_series.shape[0]
            self.sizeY = self.image_series.shape[1]
        self.frame_num = None
        self.current_frame = 0
        self.title = name
        self.mask = mask
        self.starter()
        self.save_video = None
        self.step_size = None
        self.time_delay = None
        self.fps = None
        self.info_image_save = None
        self.alpha = None
        self.beta = None
        self.max_intensity = None
        self.min_intensity = None
        self.frame_strides = 1
        self.flag_median_filter = False
        self.window_size = 1
        self.marker_size = 5
        self.flag_smooth_filter = True

        self.signals = WorkerSignals()
        self.sliceNumber.connect(self.Update_SliceNumber)

    def __del__(self):
        print("Destructor called, Employee deleted.")

    def starter(self):
        self.setWindowIcon(QtGui.QIcon(QtCore.QDir.currentPath() + "/icons/mpl.png"))
        self.setWindowTitle(self.title)
        self.setGeometry(490, 100, 512 + 124, 512 + 93)
        self.main_Viewer_frame = QtWidgets.QFrame()
        self.setCentralWidget(self.main_Viewer_frame)
        self.main_Viewer_frame_layout = QtWidgets.QGridLayout()
        self.main_Viewer_frame.setLayout(self.main_Viewer_frame_layout)
        self.resize(500, 500)

        # Create an instant of GraphicView Class to show the image PixMap.
        self.viewer = slice_view.SliceView(self.image_series, self.original_video, mask=self.mask)
        self.viewer.frameClicked.connect(self.frame_read)
        self.viewer.photoClicked.connect(self.photoClicked)

        self.main_Viewer_frame_layout.addWidget(self.viewer, 0, 0)

        # checkbox median Filter
        self.checkbox_medianFilter = QtWidgets.QRadioButton("Median Filter")
        self.checkbox_medianFilter.toggled.connect(
            lambda: self.activeMedianFilter(self.checkbox_medianFilter)
        )

        # Contrast
        self.con_adj = QtWidgets.QPushButton("Contrast adjustment")
        self.con_adj.setAutoDefault(False)
        self.con_adj.clicked.connect(self.open_adjustment)
        self.con_adj.setFixedWidth(130)
        self.con_adj.setStyleSheet("background-color : lightgrey")

        # Display
        self.Display = QtWidgets.QPushButton("Display")
        self.Display.setAutoDefault(False)
        self.Display.clicked.connect(self.matplotlib_display)
        self.Display.setFixedWidth(130)
        self.Display.setStyleSheet("background-color : lightgrey")

        # Annotation
        self.annotation_s = QtWidgets.QPushButton("Live spatial annotation")
        self.annotation_s.setCheckable(True)
        self.annotation_s.clicked.connect(self.annotation_spatial_Fun)
        self.annotation_s.setFixedWidth(150)
        self.annotation_s.setStyleSheet("background-color : lightgrey")

        self.annotation_t = QtWidgets.QPushButton("Live temporal annotation")
        self.annotation_t.setCheckable(True)
        self.annotation_t.clicked.connect(self.annotation_temporal_Fun)
        self.annotation_t.setFixedWidth(150)
        self.annotation_t.setStyleSheet("background-color : lightgrey")
        self.annotation_t.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.annotation_t.customContextMenuRequested.connect(self.annotation_context_menu)

        # Play
        self.playBtn = QtWidgets.QPushButton()
        self.playBtn.setAutoDefault(False)
        self.playBtn.setIcon(self.style().standardIcon(QtWidgets.QStyle.SP_MediaPlay))
        self.playBtn.setFixedWidth(20)
        self.playBtn.clicked.connect(self.run_play)

        # create context menu
        self.playBtn.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.playBtn.customContextMenuRequested.connect(self.play_context_menu)

        # stop
        self.stopBtn = QtWidgets.QPushButton()
        self.stopBtn.setAutoDefault(False)
        self.stopBtn.setIcon(self.style().standardIcon(QtWidgets.QStyle.SP_MediaStop))
        self.stopBtn.setFixedWidth(20)
        self.stopBtn.clicked.connect(self.stop_video)

        # Pause
        self.pauseBtn = QtWidgets.QPushButton()
        self.pauseBtn.setAutoDefault(False)
        self.pauseBtn.setIcon(self.style().standardIcon(QtWidgets.QStyle.SP_MediaPause))
        self.pauseBtn.setFixedWidth(20)
        self.pauseBtn.clicked.connect(self.pause_video)
        self._running = False

        # save
        self.save = QtWidgets.QPushButton("save")
        self.save.setAutoDefault(False)
        self.save.clicked.connect(self.do_update_save)
        self.save.setFixedWidth(130)
        self.save.setStyleSheet("background-color : lightgrey")

        if self.mask is True:
            self.Display_mask = QtWidgets.QPushButton("Display_with_mask")
            self.Display_mask.setAutoDefault(False)
            self.Display_mask.clicked.connect(self.matplotlib_display_mask)
            self.Display_mask.setFixedWidth(130)
            self.Display_mask.setStyleSheet("background-color : lightgrey")

            # mp4
            self.mp4 = QtWidgets.QPushButton("MP4+localization")
            self.mp4.setAutoDefault(False)
            self.mp4.clicked.connect(self.make_mp4)
            self.mp4.setFixedWidth(130)
            self.mp4.setStyleSheet("background-color : lightgrey")

            # progressBar
            self.statusbar = self.statusBar()
            self.progressBar = QtWidgets.QProgressBar(self)
            self.statusBar().addPermanentWidget(self.progressBar)
            self.progressBar.setGeometry(30, 40, 200, 25)
            self.progressBar.setValue(0)
            self.progressChanged.connect(self.setProgress)

        # Create
        self.connect(
            QtGui.QShortcut(QtGui.QKeySequence(QtCore.Qt.Key_H), self),
            QtCore.SIGNAL("activated()"),
            self.histogram,
        )

        if self.title != "PNG":
            self.slice_slider = QtWidgets.QSlider(QtCore.Qt.Horizontal)
            self.slice_slider.setMinimum(0)
            self.slice_slider.setMaximum(self.image_series.shape[0] - 1)
            self.slice_slider.setValue(0)
            self.slice_slider.valueChanged.connect(self.set_new_slice_num)

            self.main_Viewer_frame_layout.addWidget(self.createThirdExclusiveGroup(), 1, 0)

        self.main_Viewer_frame_layout.addWidget(self.createSecondExclusiveGroup(), 2, 0)

        self.statusbar = self.statusBar()
        self.show()

    def createSecondExclusiveGroup(self):
        self.groupBox_displayBtn = QtWidgets.QGroupBox("Video Display")

        self.grid_diplay = QtWidgets.QGridLayout()
        self.grid_diplay.addWidget(self.Display, 0, 0)
        self.grid_diplay.addWidget(self.annotation_s, 0, 1)
        self.grid_diplay.addWidget(self.con_adj, 0, 2)

        if self.mask is True:
            self.grid_diplay.addWidget(self.Display_mask, 1, 0)

        self.grid_diplay.addWidget(self.annotation_t, 1, 1)
        self.grid_diplay.addWidget(self.save, 1, 2)
        self.groupBox_displayBtn.setLayout(self.grid_diplay)

        return self.groupBox_displayBtn

    def createThirdExclusiveGroup(self):
        play = QtWidgets.QGroupBox()

        self.grid_play = QtWidgets.QGridLayout()
        self.grid_play.addWidget(self.playBtn, 0, 0)
        self.grid_play.addWidget(self.pauseBtn, 0, 1)
        self.grid_play.addWidget(self.stopBtn, 0, 2)
        self.grid_play.addWidget(self.slice_slider, 0, 3)
        self.grid_play.addWidget(self.checkbox_medianFilter, 0, 4)

        play.setLayout(self.grid_play)
        return play

    def activeMedianFilter(self, box_status):
        if box_status.isChecked():
            self.viewer.medianFilterFlag = True
            self.flag_median_filter = True
        else:
            self.viewer.medianFilterFlag = False
            self.flag_median_filter = False

    def annotation_spatial_Fun(self):
        if self.annotation_s.isChecked():
            self.annotation_s.setStyleSheet("background-color : lightblue")

        else:
            self.annotation_s.setStyleSheet("background-color : lightgrey")

        self.viewer.annotation_s.emit()

    def annotation_temporal_Fun(self):
        if self.annotation_t.isChecked():
            self.annotation_t.setStyleSheet("background-color : lightblue")

        else:
            self.annotation_t.setStyleSheet("background-color : lightgrey")

        self.viewer.annotation_t.emit(
            [self.window_size, self.flag_smooth_filter, self.marker_size]
        )

    def closeEvent(self, event):
        try:
            self.stop_video()
        except:
            pass
        print("closed")

    @QtCore.Slot()
    def breakAll(self):
        self.make_mp4.kill()

    @QtCore.Slot()
    def Update_SliceNumber(self, frame_number):
        self.current_frame = frame_number
        self.currentFrame.emit(frame_number)

    def get_in(self, df_PSFs):
        self.df_PSFs = df_PSFs

    @QtCore.Slot()
    def call_save(self):
        if self.fps == "" or self.fps is None:
            self.fps = 10
        else:
            self.fps = int(self.fps)

        if self.info_image_save.video_type == "RAW":
            if self.info_image_save.bin_type is not None:
                type = self.info_image_save.bin_type
            else:
                type = "original"

            file_name = (
                str(self.original_video.shape[0])
                + "_"
                + str(self.original_video.shape[1])
                + "_"
                + str(self.original_video.shape[2])
                + "_"
                + str(type)
                + ".raw"
            )
            write_video.write_binary(
                dir_path=self.file_path, file_name=file_name, data=self.original_video, type=type
            )

        elif self.info_image_save.video_type == "MP4":
            self.flag_update_normalization = True
            worker = Normalization(video=self.original_video, flag_image_specific=True)
            worker.signals.result.connect(self.update_normalization)
            self.threadpool.start(worker)
            while self.flag_update_normalization:
                QtCore.QCoreApplication.processEvents()
            file_name = (
                str(self.original_video.shape[0])
                + "_"
                + str(self.original_video.shape[1])
                + "_"
                + str(self.original_video.shape[2])
                + "_"
                + str(self.original_video.dtype)
                + ".mp4"
            )
            write_video.write_MP4(
                dir_path=self.file_path,
                file_name=file_name,
                data=self.save_video,
                jump=self.frame_strides,
                fps=self.fps,
            )

        elif self.info_image_save.video_type == "GIF":
            self.flag_update_normalization = True
            worker = Normalization(video=self.original_video, flag_image_specific=True)
            worker.signals.result.connect(self.update_normalization)
            self.threadpool.start(worker)
            while self.flag_update_normalization:
                QtCore.QCoreApplication.processEvents()
            file_name = (
                str(self.original_video.shape[0])
                + "_"
                + str(self.original_video.shape[1])
                + "_"
                + str(self.original_video.shape[2])
                + "_"
                + str(self.original_video.dtype)
                + ".gif"
            )

            write_video.write_GIF(
                dir_path=self.file_path,
                file_name=file_name,
                data=self.save_video,
                jump=self.frame_strides,
                duration=1000/self.fps,
            )

    @Slot()
    def update_normalization(self, video):
        self.save_video = video
        self.flag_update_normalization = False

    def do_update_save(self):
        if self.save.clicked:
            try:
                self.info_image_save = save_GUI.SaveVideo()
                self.info_image_save.window.show()

                self.info_image_save.signal_save_Done.connect(self.call_save)
                self.file_path = self.askdir()

            except:
                self.msg_box = QtWidgets.QMessageBox()
                self.msg_box.setWindowTitle("Warning!")
                self.msg_box.setText("Folder does not defined!")
                self.msg_box.exec_()

    def askdir(self):
        folder = QtWidgets.QFileDialog.getExistingDirectory(
            None, "Select a folder:", "", QtWidgets.QFileDialog.ShowDirsOnly
        )
        return folder

    def matplotlib_display(self):
        if self.time_delay == "" or self.time_delay is None:
            self.time_delay = 0.1
        else:
            self.time_delay = float(self.time_delay)

        if self.Display.clicked:
            if self.step_size != "" and self.step_size is not None:
                Display(
                    video=np.fliplr(self.original_video),
                    time_delay=float(self.time_delay),
                    median_filter_flag=self.flag_median_filter,
                )
            else:
                Display(
                    video=np.fliplr(self.original_video),
                    time_delay=float(self.time_delay),
                    median_filter_flag=self.flag_median_filter,
                )

    def matplotlib_display_mask(self):
        if self.step_size == "" or self.step_size is None:
            self.step_size = 1
        else:
            self.step_size = int(self.step_size)

        if self.time_delay == "" or self.time_delay is None:
            self.time_delay = 0.1
        else:
            self.time_delay = float(self.time_delay)

        if self.Display_mask.clicked:
            if self.df_PSFs is not None:
                if type(self.df_PSFs) is pd.core.frame.DataFrame:
                    if self.step_size != "":
                        ani_ = DisplayPSFs_subplotLocalizationDisplay(
                            list_videos=[self.original_video],
                            list_df_PSFs=[self.df_PSFs],
                            list_titles=None,
                            numRows=1,
                            numColumns=1,
                            color="gray",
                            median_filter_flag=self.viewer.medianFilterFlag,
                            imgSizex=10,
                            imgSizey=5,
                            time_delay=self.time_delay,
                        )

                    else:
                        ani_ = DisplayPSFs_subplotLocalizationDisplay(
                            list_videos=[self.original_video],
                            list_df_PSFs=[self.df_PSFs],
                            list_titles=None,
                            numRows=1,
                            numColumns=1,
                            color="gray",
                            median_filter_flag=self.viewer.medianFilterFlag,
                            imgSizex=10,
                            imgSizey=5,
                            time_delay=self.time_delay,
                        )

            else:
                self.msg_box = QtWidgets.QMessageBox()
                self.msg_box.setWindowTitle("Warning!")
                self.msg_box.setText("The PSF location is not defined")
                self.msg_box.exec_()

    def make_mp4(self):
        if self.mp4.clicked:
            self.file_path = self.askdir()
            try:
                if self.step_size == "" or self.step_size is None:
                    self.step_size = 1
                else:
                    self.step_size = int(self.step_size)

                if self.time_delay == "" or self.time_delay is None:
                    self.time_delay = 0.1
                else:
                    self.time_delay = float(self.time_delay)

                if self.fps == "" or self.fps is None:
                    self.fps = 10
                else:
                    self.fps = int(self.fps)

                p_max = self.original_video.shape[0] - self.step_size
                self.progressBar.setValue(0)
                self.progressBar.setMaximum(p_max)

                if self.df_PSFs is not None:
                    if type(self.df_PSFs) is pd.core.frame.DataFrame:
                        self.flag_mp4 = True
                        save_path = os.path.join(self.file_path, "PiSCAT_GUI_Localization.mp4")

                        ani_ = HTML_PSFs_subplotLocalizationDisplay(
                            list_videos=[self.original_video],
                            list_df_PSFs=[self.df_PSFs],
                            list_titles=None,
                            numRows=1,
                            numColumns=1,
                            cmap="gray",
                            median_filter_flag=self.viewer.medianFilterFlag,
                            imgSizex=10,
                            imgSizey=5,
                            time_delay=self.time_delay,
                            save_path=save_path,
                            fps=self.fps,
                        )
                        ani_.signals.finished.connect(self.finish_mp4)
                        self.threadpool.start(ani_)
                        while self.flag_mp4:
                            QtCore.QCoreApplication.processEvents()

                        self.msg_box = QtWidgets.QMessageBox()
                        self.msg_box.setWindowTitle("Warning!")
                        self.msg_box.setText("MP4 vide saved!")
                        self.msg_box.exec_()

                else:
                    self.msg_box = QtWidgets.QMessageBox()
                    self.msg_box.setWindowTitle("Warning!")
                    self.msg_box.setText("The PSF location is not defined")
                    self.msg_box.exec_()
            except:
                self.msg_box = QtWidgets.QMessageBox()
                self.msg_box.setWindowTitle("Warning!")
                self.msg_box.setText("Error occurred in MP4!")
                self.msg_box.exec_()

    @QtCore.Slot()
    def finish_mp4(self):
        self.flag_mp4 = False

    def startProgressBar(self, fun, **kwargs):
        self.thread = fun(**kwargs)
        self.thread.obj_connection.updateProgress_.connect(self.setProgress)
        self.thread.obj_connection.finished.connect(self.thread_complete)
        self.threadpool.start(self.thread)

    @QtCore.Slot()
    def update_spatial_animate(self, setting):
        self.step_size = setting[0]
        self.time_delay = setting[1]
        self.fps = setting[2]
        self.raise_()

    @QtCore.Slot()
    def update_temporal_annotation(self, setting):
        self.window_size = setting[0]
        self.flag_smooth_filter = setting[1]
        self.marker_size = setting[2]
        self.raise_()

    def set_new_slice_num(self):
        self.viewer.slice_num = self.slice_slider.value()
        self.updata_viewer()
        self.sliceNumber.emit(self.viewer.slice_num)

    def enterEvent(self, event):
        self.setMouseTracking(True)

    def leaveEvent(self, event):
        self.setMouseTracking(False)

    def update_potion(self, result_):
        self.statusbar.showMessage(result_)

    def thread_complete(self):
        print("THREAD COMPLETE!")

    def wheelEvent(self, *args, **kwargs):
        pass

    def histogram(self):
        tmp = self.original_video[self.viewer.slice_num]
        plt.figure(figsize=(6, 6), dpi=100)
        plt.hist(tmp.ravel(), fc="g", ec="k")
        plt.title("Histogram of pixel intensity")
        plt.ylabel("#Counts")
        plt.show()

    @QtCore.Slot()
    def run_play(self):
        if self.step_size == "" or self.step_size is None:
            self.frame_strides = 1
        else:
            self.frame_strides = int(self.step_size)

        if self.time_delay == "" or self.time_delay is None:
            self.time_delay = 0.1
        else:
            self.time_delay = float(self.time_delay)

        self.playBtn.setDisabled(True)
        self.play_video()

    @QtCore.Slot()
    def play_video(self):
        global play_flag
        play_flag = True
        slice_num = self.slice_slider.value()
        self.playBtn.setDisabled(True)

        while slice_num < self.image_series.shape[0] - 1 and play_flag:
            slice_num += self.frame_strides
            self.slice_slider.setValue(slice_num)
            self.viewer.slice_num = self.slice_slider.value()
            self.updata_viewer()
            QtCore.QCoreApplication.processEvents()

    @QtCore.Slot()
    def pause_video(self):
        global play_flag
        play_flag = False
        self.playBtn.setEnabled(True)

    @QtCore.Slot()
    def stop_video(self):
        global play_flag
        play_flag = False

        self.playBtn.setEnabled(True)

        self.slice_slider.setValue(0)
        self.viewer.slice_num = self.slice_slider.value()
        self.updata_viewer()

    def play_context_menu(self):
        self.animatePanel_ = PlaySetting()
        self.animatePanel_.procDone_Animation_panel.connect(self.update_spatial_animate)
        self.animatePanel_.show()

    def annotation_context_menu(self):
        self.temporal_annotation = TemporalAnnotationSetting()
        self.temporal_annotation.temporalAnnotation_signal.connect(
            self.update_temporal_annotation
        )
        self.temporal_annotation.show()

    def frame_read(self, frame_num):
        if self.viewer.dragMode() == QtWidgets.QGraphicsView.NoDrag:
            self.frame_num = frame_num

    def photoClicked(self, pos):
        if self.viewer.dragMode() == QtWidgets.QGraphicsView.NoDrag:
            value = self.original_video[self.frame_num, pos[0], pos[1]]
            self.statusbar.showMessage(
                "X:%d, Y:%d, Frame:%d, Value:%f" % (pos[0], pos[1], self.frame_num, value)
            )

    @QtCore.Slot(int)
    def setProgress(self, val):
        self.progressBar.setValue(val)

    @QtCore.Slot()
    def set_alpha(self, alpha):
        self.alpha = alpha

    @QtCore.Slot()
    def set_beta(self, beta):
        self.beta = beta

    @QtCore.Slot()
    def set_max_intensity(self, max_intensity):
        self.max_intensity = max_intensity
        self.updata_viewer()

    @QtCore.Slot()
    def set_min_intensity(self, min_intensity):
        self.min_intensity = min_intensity

    def open_adjustment(self):
        img_ = self.viewer.input_file
        min_intensity = img_.min()
        max_intensity = img_.max()
        alpha = 255 / (max_intensity - min_intensity)
        beta = -min_intensity * alpha
        self.con_adj_setting = Contrast_adjustment_GUI(alpha, beta, min_intensity, max_intensity)

        self.con_adj_setting.alpha_signal.connect(self.set_alpha)
        self.con_adj_setting.beta_signal.connect(self.set_beta)
        self.con_adj_setting.min_intensity_signal.connect(self.set_min_intensity)
        self.con_adj_setting.max_intensity_signal.connect(self.set_max_intensity)

    def updata_viewer(self):
        self.viewer.current_pixmap = self.viewer.create_pixmap(
            self.image_series[self.viewer.slice_num]
        )
        self.viewer.update_slice(self.viewer.current_pixmap)
        if self.viewer.mask_is_set is True:
            self.viewer.current_mask_pixmap = self.viewer.create_mask_pixmap(
                self.viewer.maskArray[self.viewer.slice_num, :, :]
            )
            self.viewer.update_overlay(self.viewer.current_mask_pixmap)



import numpy as np
import pandas as pd


def feature2df(feature_position, videos):
    if feature_position.shape[1] == 4:
        sigma = feature_position[:, 3]
        psf_position_x = feature_position[:, 2]
        psf_position_y = feature_position[:, 1]
        psf_position_frame = np.asarray(feature_position[:, 0], dtype=int)

        center_intensity = np.asarray(
            [
                videos[
                    int(psf_position_frame[i_]), int(psf_position_y[i_]), int(psf_position_x[i_])
                ]
                for i_ in range(psf_position_frame.shape[0])
            ]
        )

        dict = {
            "y": psf_position_y,
            "x": psf_position_x,
            "frame": psf_position_frame,
            "center_intensity": center_intensity,
            "sigma": sigma,
        }

        df_features = pd.DataFrame(dict)

        return df_features

    elif feature_position.shape[1] == 5:
        sigma = (feature_position[:, 3], feature_position[:, 4])
        sigma = np.asarray(sigma)
        psf_position_x = feature_position[:, 2]
        psf_position_y = feature_position[:, 1]
        psf_position_frame = np.asarray(feature_position[:, 0], dtype=np.int64)

        center_intensity = np.asarray(
            [
                videos[
                    int(psf_position_frame[i_]), int(psf_position_y[i_]), int(psf_position_x[i_])
                ]
                for i_ in range(psf_position_frame.shape[0])
            ]
        )

        dict = {
            "y": psf_position_y,
            "x": psf_position_x,
            "frame": psf_position_frame,
            "center_intensity": center_intensity,
            "sigma_x": feature_position[:, 3],
            "sigma_y": feature_position[:, 4],
        }

        df_features = pd.DataFrame(dict)
        df_features["sigma"] = 0.5 * (df_features["sigma_x"] + df_features["sigma_y"])
        return df_features


def list2dataframe(feature_position, video):
    """This function converts the output of
    ``particle_localization.PSFsExtraction`` method from list to data frame.

    Parameters
    ----------
    feature_position: list
        List of position of PSFs (x, y, frame, sigma)

    video: NDArray
        The video is 3D-numpy (number of frames, width, height).

    Returns
    -------
    df_features: pandas dataframe
        PSF positions are stored in the data frame. ( 'y', 'x', 'frame',
        'center_intensity', 'sigma', 'Sigma_ratio', ...).

    """
    if feature_position is not None:
        if type(feature_position) is list:
            if len(feature_position) == 0:
                raise Exception("---List is empty!---")
            feature_position = [l_ for l_ in feature_position if l_ is not None]
            if len(feature_position) >= 1:
                feature_position = np.concatenate(feature_position, axis=0)
                df_features = feature2df(feature_position, video)
            else:
                df_features = None

        elif type(feature_position) is np.ndarray:
            df_features = feature2df(feature_position, video)
        else:
            df_features = None

        return df_features
    else:
        df_features = None

        return df_features



from PySide6 import QtCore, QtWidgets
from PySide6.QtCore import QThreadPool

from piscat.GUI.Visualization.fun_display_localization import Visulization_localization
from piscat.Preproccessing.filtering import FFT2D


class FFT2D_GUI_wrapper:
    def __init__(self):
        super(FFT2D_GUI_wrapper, self).__init__()
        self.flag_update_FFT_video = True
        self.original_video = None
        self.fft_v = None
        self.threadpool = QThreadPool()

    def spectrum_input(self, video):
        self.original_video = video

    def spectrum(self):
        if self.original_video is not None:
            self.flag_update_FFT_video = True
            worker = FFT2D(video=self.original_video)
            worker.signals.result.connect(self.update_FFT_video)
            self.threadpool.start(worker)
            while self.flag_update_FFT_video:
                QtCore.QCoreApplication.processEvents()

            self.flag_update_FFT_video = True

            visualization_ = Visulization_localization()
            visualization_.new_display(self.fft_v, self.fft_v, object=None, title="Spectrum")

        else:
            self.msg_box = QtWidgets.QMessageBox()
            self.msg_box.setWindowTitle("Warning!")
            self.msg_box.setText("Please load one video!")
            self.msg_box.exec_()

    def update_FFT_video(self, r_):
        self.fft_v = r_
        self.flag_update_FFT_video = False



import math

import numpy as np

from piscat.iPSF_model.VectorialDiffractionIntegral import VectorialDiffractionIntegral


class ScatteredFieldDifferentialPhase:
    def __init__(self, p, Xp, z_focus, nz, nx):
        """To calculate the Electric field scattered from a nanoparticle using
        a vectorial diffraction formulation.

        References
        ----------
        "Point spread function in interferometric scattering microscopy
        (iSCAT). Part I: aberrations in defocusing and axial localization.",
        Reza Gholami Mahmoodabadi, Richard W. Taylor, Martin Kaller, Susann
        Spindler, Mahdi Mazaheri, Kiarash Kasaian, and Vahid Sandoghdar Optics
        Express Vol. 28, Issue 18, pp. 25969-25988 (2020)
        https://doi.org/10.1364/OE.401374

        Parameters
        ----------
        p : Imaging setup parameters which is an object from the class imagingSetupParameters
        Xp : The 3D position of the scatterer in the sample, in unit of meters
        z_focus : The position of the focus. This could be an array as well.
        nz : The number of focus positions to calculate.
        nx : The lateral extent of the image given in pixels

        """
        self.p = p
        self.Xp = Xp
        self.z_focus = z_focus
        self.nz = nz
        self.nx = nx

        pass

    def calculate(self, r_):
        p = self.p
        Xp = self.Xp
        z_focus = self.z_focus
        nz = self.nz
        nx = self.nx

        vec_diffraction_integral = VectorialDiffractionIntegral(p, Xp, z_focus, nz, nx)
        vec_diffraction_integral.calculate()

        I0_re = vec_diffraction_integral.I0_int_re_op_3D
        I1_re = vec_diffraction_integral.I1_int_re_op_3D
        I2_re = vec_diffraction_integral.I2_int_re_op_3D
        I0_im = vec_diffraction_integral.I0_int_im_op_3D
        I1_im = vec_diffraction_integral.I1_int_im_op_3D
        I2_im = vec_diffraction_integral.I2_int_im_op_3D

        I0 = I0_re + 1j * I0_im
        I1 = I1_re + 1j * I1_im
        I2 = I2_re + 1j * I2_im

        nxi = np.shape(I0)[2]  # size(I0,2)
        xpMesh = (nxi + 1) / 2
        nyi = np.shape(I0)[1]  # size(I0,1)
        ypMesh = (nyi + 1) / 2

        Y, X = np.meshgrid(np.arange(1, 1 + nyi), np.arange(1, 1 + nxi))
        pixel_size = p.pixel_size_physical / p.M
        xp_ = Xp[0]
        yp_ = Xp[1]
        Xnm = (X - xpMesh) * pixel_size - xp_
        Ynm = (Y - ypMesh) * pixel_size - yp_
        eps = np.finfo(float).eps
        PhiD_ = np.arctan(Xnm / (Ynm + eps))
        PhiD = np.tile(PhiD_, (np.shape(I0_re)[0], 1, 1))  # size(I0_re,3))
        # EFTppol = -1j *(I0 +  I2 .* cos(2*PhiD));
        # xpread = round(xpMesh+xp_/pixel_size)
        # ypread = round(ypMesh+yp_/pixel_size)
        xpread = round(xpMesh)
        ypread = round(ypMesh)

        M11 = I0 + I2 * np.cos(2 * PhiD)
        M12 = I2 * np.sin(2 * PhiD)
        M13 = -2j * I1 * np.cos(PhiD)

        # M21 = (I2 .* sin(2*PhiD));
        # M22 = (I0 -  I2 .* cos(2*PhiD));
        # M23 = -2i*I1.*sin(PhiD);

        # dipole orientation
        phiP = 0  # azimuth
        thetP = math.pi / 2  # zenith
        P11 = np.sin(thetP) * np.cos(phiP)
        P21 = np.sin(thetP) * math.sin(phiP)
        P31 = np.cos(thetP)
        EE11 = -1j * (M11 * P11 + M12 * P21 + M13 * P31)
        # EE21 = -1j*(M21.*P11 + M22.*P21 + M23.*P31);

        angsEE11 = np.angle(
            np.squeeze(
                EE11[:, (xpread - r_ - 1) : (xpread + r_), (ypread - r_ - 1) : (ypread + r_)]
            )
        )
        ampsEE11 = np.abs(
            np.squeeze(
                EE11[:, (xpread - r_ - 1) : (xpread + r_), (ypread - r_ - 1) : (ypread + r_)]
            )
        )

        return ampsEE11, angsEE11



import os

import numpy as np

from piscat.BackgroundCorrection import DRA
from piscat.InputOutput import read_status_line, read_write_data, reading_videos
from piscat.Localization import localization_filtering, particle_localization
from piscat.Preproccessing import filtering, normalization
from piscat.Trajectory import particle_linking, temporal_filtering


def protein_analysis(paths, video_names, hyperparameters, flags, name_mkdir):
    """This function analyses several videos based on the setting that the
    user defines in the ``hyperparameters`` and ``flags``.

    Parameters
    ----------
    paths: list
        List of all paths for videos.

    video_names: str
        list of all video names.

    hyperparameters: dic
        The dictionary is used to define different parameters for analysis. In
        the following you can see the example of this dictionary:

            | hyperparameters = {'function': 'dog', 'batch_size': 3000,
                                 'min_V_shape_width': 1500, 'threshold_max':
                                 6000 'search_range': 2, 'memory': 20,
                                 'min_sigma': 1.3, 'max_sigma': 3,
                                 'sigma_ratio': 1.1, 'PSF_detection_thr': 4e-5,
                                 'overlap': 0, 'outlier_frames_thr': 20,
                                 'Mode_PSF_Segmentation': 'BOTH',
                                 'symmetric_PSFs_thr': 0.6, 'mode_FPN':
                                 name_mkdir, 'select_correction_axis': 1,
                                 'im_size_x': 72, 'im_size_y': 72,
                                 'image_format': '<u2'}

    flags: dic
         The dictionary is used to active/deactivate different parts in
         analyzing pipelines. In the following you can see the example of this
         dictionary:

            | flags = {'PN': True, 'FPNc': True, 'outlier_frames_filter': True,
              'Dense_Filter': True, 'symmetric_PSFs_Filter': True, 'FFT_flag':
              True}

    name_mkdir: str
        It defines the name of the folder that automatically creates next to
        each video to save the results of the analysis and setting history.

    Returns
    -------
    The following information will be saved

        * `hyperparameters`

        * `flags`

        * `Number of particles and PSFs after each steps`

        * `All extracted trajectories with 'HDF5' and 'Matlab' format.`

            * `MATLAB` saves array contains the following information for each
              particle:

            | [intensity_horizontal, intensity_vertical,
               particle_center_intensity, particle_center_intensity_follow,
               particle_frame, particle_sigma, particle_X, particle_Y,
               particle_ID, optional(fit_intensity, fit_x, fit_y, fit_X_sigma,
               fit_Y_sigma, fit_Bias, fit_intensity_error, fit_x_error,
               fit_y_error, fit_X_sigma_error, fit_Y_sigma_error,
               fit_Bias_error)]

            * `HDF5` saves dictionary similar to the following structures:

            | {"#0": {'intensity_horizontal': ..., 'intensity_vertical': ...,
                        ..., 'particle_ID': ...}, "#1": {}, ...}

        * `Table of PSFs information`

            * `particles`: pandas dataframe

            | Saving the clean data frame (x, y, frame, sigma, particle, ...)

    """
    PSFs_Particels_num = {
        "#Totall_PSFs": None,
        "#PSFs_afterOutlierFramesFilter": None,
        "#PSFs_afterDenseFilter": None,
        "#PSFs_afterSymmetricPSFsFilter": None,
        "#Totall_Particles": None,
        "#Particles_after_V_shapeFilter": None,
        "#Totall_frame_num_DRA": None,
    }

    for p_, n_ in zip(paths, video_names):
        print("---" + p_ + "---")
        try:
            dr_mk = os.path.join(p_, name_mkdir)
            os.mkdir(dr_mk)
            print("Directory ", name_mkdir, " Created ")
        except FileExistsError:
            print("Directory ", name_mkdir, " already exists")

        s_dir_ = os.path.join(p_, name_mkdir)
        read_write_data.save_dic2json(
            data_dictionary=hyperparameters, path=s_dir_, name="hyperparameters"
        )
        read_write_data.save_dic2json(data_dictionary=flags, path=s_dir_, name="flags")

        video = reading_videos.read_binary(
            file_name=p_ + "/" + n_,
            img_width=hyperparameters["im_size_x"],
            img_height=hyperparameters["im_size_y"],
            image_type=np.dtype(hyperparameters["image_format"]),
        )

        if isinstance(hyperparameters["start_fr"], int) and isinstance(
            hyperparameters["end_fr"], int
        ):
            video = video[hyperparameters["start_fr"] : hyperparameters["end_fr"], :, :]

        status_ = read_status_line.StatusLine(video)
        video, status_info = status_.find_status_line()

        if flags["PN"]:
            video_pn, _ = normalization.Normalization(video=video).power_normalized()
        else:
            video_pn = video

        DRA_ = DRA.DifferentialRollingAverage(
            video=video_pn,
            batchSize=hyperparameters["batch_size"],
            mode_FPN=hyperparameters["mode_FPN"],
        )

        RVideo_PN_FPN_, _ = DRA_.differential_rolling(
            FPN_flag=flags["FPNc"],
            select_correction_axis=hyperparameters["select_correction_axis"],
            FFT_flag=flags["FFT_flag"],
            inter_flag_parallel_active=True,
        )

        if flags["filter_hotPixels"]:
            RVideo_PN_FPN = filtering.Filters(RVideo_PN_FPN_).median(3)
        else:
            RVideo_PN_FPN = RVideo_PN_FPN_

        PSFs_Particels_num["#Totall_frame_num_DRA"] = RVideo_PN_FPN.shape[0]

        PSF_l = particle_localization.PSFsExtraction(video=RVideo_PN_FPN)
        df_PSFs = PSF_l.psf_detection(
            function=hyperparameters["function"],  # function='log', 'doh', 'dog'
            min_sigma=hyperparameters["min_sigma"],
            max_sigma=hyperparameters["max_sigma"],
            sigma_ratio=hyperparameters["sigma_ratio"],
            threshold=hyperparameters["PSF_detection_thr"],
            overlap=hyperparameters["overlap"],
            mode=hyperparameters["Mode_PSF_Segmentation"],
        )

        if df_PSFs is not None:
            PSFs_Particels_num["#Totall_PSFs"] = df_PSFs.shape[0]
        else:
            PSFs_Particels_num["#Totall_PSFs"] = 0

        if PSFs_Particels_num["#Totall_PSFs"] > 1:
            s_filters = localization_filtering.SpatialFilter()

            if flags["outlier_frames_filter"]:
                df_PSFs_s_filter = s_filters.outlier_frames(
                    df_PSFs, threshold=hyperparameters["outlier_frames_thr"]
                )
                PSFs_Particels_num["#PSFs_afterOutlierFramesFilter"] = df_PSFs_s_filter.shape[0]
            else:
                df_PSFs_s_filter = df_PSFs

            if flags["Dense_Filter"]:
                df_PSFs_s_filter = s_filters.dense_PSFs(df_PSFs_s_filter, threshold=0)
                PSFs_Particels_num["#PSFs_afterDenseFilter"] = df_PSFs_s_filter.shape[0]

            if flags["symmetric_PSFs_Filter"] and df_PSFs_s_filter.shape[0] != 0:
                df_PSF_2Dfit = PSF_l.fit_Gaussian2D_wrapper(
                    PSF_List=df_PSFs, scale=5, internal_parallel_flag=True
                )
                df_PSFs_s_filter = s_filters.symmetric_PSFs(
                    df_PSFs=df_PSF_2Dfit, threshold=hyperparameters["symmetric_PSFs_thr"]
                )
                if df_PSFs_s_filter is not None:
                    PSFs_Particels_num["#PSFs_afterSymmetricPSFsFilter"] = df_PSFs_s_filter.shape[
                        0
                    ]

            if df_PSFs_s_filter.shape[0] > 1 and df_PSFs_s_filter is not None:
                linking_ = particle_linking.Linking()
                df_PSFs_link = linking_.create_link(
                    psf_position=df_PSFs_s_filter,
                    search_range=hyperparameters["search_range"],
                    memory=hyperparameters["memory"],
                )
                PSFs_Particels_num["#Totall_Particles"] = linking_.trajectory_counter(
                    df_PSFs_link
                )

                t_filters = temporal_filtering.TemporalFilter(
                    video=RVideo_PN_FPN, batchSize=hyperparameters["batch_size"]
                )
                all_trajectories, df_PSFs_t_filter, his_all_particles = t_filters.v_trajectory(
                    df_PSFs=df_PSFs_link,
                    threshold_min=hyperparameters["min_V_shape_width"],
                    threshold_max=hyperparameters["max_V_shape_width"],
                )

                PSFs_Particels_num[
                    "#Particles_after_V_shapeFilter"
                ] = linking_.trajectory_counter(df_PSFs_t_filter)

                read_write_data.save_df2csv(df_PSFs_t_filter, path=s_dir_, name="position_PSFs")

        read_write_data.save_dic2json(
            data_dictionary=PSFs_Particels_num, path=s_dir_, name="PSFs_Particels_num"
        )



from .camera_setting import CameraParameters
from .cpu_configurations import CPUConfigurations
from .image_to_video import Image2Video
from .read_status_line import StatusLine
from .read_write_data import (
    download_tutorial_data,
    load_dict_from_hdf5,
    read_json2dic,
    read_mat,
    save_df2csv,
    save_dic2json,
    save_dic_to_hdf5,
    save_list_to_hdf5,
    save_mat,
)
from .reading_videos import *
from .reading_videos import (
    DirectoryType,
    read_avi,
    read_binary,
    read_fits,
    read_png,
    read_tif,
    video_reader,
)
from .write_video import write_binary, write_GIF, write_MP4

__all__ = [
    "CameraParameters",
    "CPUConfigurations",
    "Image2Video",
    "StatusLine",
    "download_tutorial_data",
    "load_dict_from_hdf5",
    "read_json2dic",
    "read_mat",
    "save_df2csv",
    "save_dic2json",
    "save_dic_to_hdf5",
    "save_list_to_hdf5",
    "save_mat",
    "DirectoryType",
    "read_avi",
    "read_binary",
    "read_fits",
    "read_png",
    "read_tif",
    "video_reader",
    "write_binary",
    "write_GIF",
    "write_MP4",
]



import os

import pandas as pd

from piscat.Analysis.plot_protein_histogram import PlotProteinHistogram
from piscat.InputOutput import read_write_data, reading_videos
from piscat.InputOutput.camera_setting import CameraParameters
from piscat.InputOutput.read_write_data import load_dict_from_hdf5


class ReadProteinAnalysis(CameraParameters):
    def __init__(self, camera_Name="Photonfocus.json"):
        """This class is developed to read the results of different video
        analyses that were analyzed with the ``protein_analysis`` function. In
        the end, this class has different methods to display histograms or save
        histogram results.

        Parameters
        ----------
        camera_Name: str
            Define the name of the camera configuration JSON file that will use
            for reading pixel size.

        """
        CameraParameters.__init__(self, name=camera_Name)

    def __call__(
        self,
        dirName,
        name_dir,
        video_frame_num,
        MinPeakWidth,
        MinPeakProminence=0,
        type_file="h5",
        his_setting=None,
    ):
        """By calling the object of class this function tries to read the
        result of the corresponding video it is defined with ``dirName`` and
        ``name_dir``. These results concatenated with previous results to use
        for plotting histogram.

        Parameters
        ----------
        dirName: str
            Path to the result of video analysis.

        name_dir: str
            The name of a folder that analysis of video was saved on it.

        video_frame_num: int
            The number of frames for the corresponding video.

        MinPeakWidth: int
            This is defined as the minimum V-shaped mouth that will use for
            prominence.

        MinPeakProminence: int
            This is defined as the minimum V-shape height that will use for
            prominence.

        type_file: str
            It defines the format of the file as the save file for analysis
            data ('HDF5', 'Matlab').

        his_setting: dict
            The dictionary is used to establish various parameters for
            localization-based filtering of PSF information in the
            histogram. The following gives an example of how this dictionary
            might be used:

            | his_setting = {'radius': 20, 'flag_localization_filter': True,
                             'centerOfImage_X': 34, 'centerOfImage_Y': 34}

        """

        df_histogram = reading_videos.DirectoryType(dirName, type_file=type_file).return_df()
        paths = df_histogram["Directory"].tolist()
        file_names = df_histogram["File"].tolist()

        self.his_ = PlotProteinHistogram(intersection_display_flag=False)
        if his_setting is None:
            self.his_.get_setting(
                radius=None,
                flag_localization_filter=False,
                centerOfImage_X=None,
                centerOfImage_Y=None,
            )
        else:
            self.his_.get_setting(
                radius=his_setting["radius"],
                flag_localization_filter=his_setting["flag_localization_filter"],
                centerOfImage_X=his_setting["centerOfImage_X"],
                centerOfImage_Y=his_setting["centerOfImage_X"],
            )

        self.dic_video_num_particle = {"Folder_name": [], "num_particles": []}

        num_particles = 0
        folder_cnt_ = 0

        for p_, n_ in zip(paths, file_names):
            if name_dir in p_:
                print(p_)
                hyperparameters = None
                PSFs_Particels_num = None
                df_json = reading_videos.DirectoryType(p_, type_file=".json").return_df()
                df_json["num_particles"] = None
                paths_json = df_json["Directory"].tolist()
                names_json = df_json["File"].tolist()
                for p_json, n_json in zip(paths_json, names_json):
                    if "flags" in n_json:
                        flags = read_write_data.read_json2dic(path=p_json, name=n_json)

                    elif "hyperparameters" in n_json:
                        hyperparameters = read_write_data.read_json2dic(path=p_json, name=n_json)

                    elif "PSFs_Particels_num" in n_json:
                        PSFs_Particels_num = read_write_data.read_json2dic(
                            path=p_json, name=n_json
                        )
                        num_particles = (
                            num_particles + PSFs_Particels_num["#Particles_after_V_shapeFilter"]
                        )
                        self.dic_video_num_particle["Folder_name"].append(p_)
                        self.dic_video_num_particle["num_particles"].append(
                            PSFs_Particels_num["#Particles_after_V_shapeFilter"]
                        )

                if hyperparameters is not None:
                    filename = os.path.join(p_, n_)
                    data_dic = load_dict_from_hdf5(filename)
                    if PSFs_Particels_num is not None:
                        if "#Totall_frame_num_DRA" in PSFs_Particels_num.keys():
                            video_frame_num = PSFs_Particels_num["#Totall_frame_num_DRA"]
                        else:
                            video_frame_num = video_frame_num

                        self.his_(
                            folder_name=n_,
                            particles=data_dic,
                            batch_size=hyperparameters["batch_size"],
                            video_frame_num=video_frame_num,
                            MinPeakWidth=MinPeakWidth,
                            MinPeakProminence=MinPeakProminence,
                            pixel_size=self.pixelSize,
                        )

                        folder_cnt_ += 1

        print("{} folders was read".format(folder_cnt_))
        print("{} PSFs discovered prior to trimming ".format(num_particles))

    def plot_localization_heatmap(
        self, pixelSize=None, unit="um", flag_in_time=False, time_delay=0.1, dir_name=None
    ):
        """This method plots heatmap of particle localization. The size of
        each disk depicts the movment of each particles during tracking.

        Parameters
        ----------
        pixelSize: float
            Camera pixel size.

        unit: str
            The axis unit.

        flag_in_time: bool
            In the case of True, show binding and unbinding events in time.

        time_delay: float
            Define the time delay between binding and unbinding events
            frames. This only works when `flag_in_time` is set to True.

        dir_name: str
            You can save time slap frames if you specify a save path.
        """
        if pixelSize is not None:
            self.pixelSize = pixelSize
        self.his_.plot_localization_heatmap(
            self.pixelSize,
            unit=unit,
            flag_in_time=flag_in_time,
            time_delay=time_delay,
            dir_name=dir_name,
        )

    def plot_hist(self, his_setting):
        """This method plots histograms for different contrast extraction
        methods for black PSFs, white PSFs and all together.

        Parameters
        ----------
        his_setting: dic
             This dictionary defines a histogram plotting setting. In the
             following you can see the example for it:

                | his_setting = {'bins': None, 'lower_limitation': -7e-4,
                                 'upper_limitation': 7e-4, 'Flag_GMM_fit':
                                 True, 'max_n_components': 3, 'step_range':
                                 1e-6, 'face': 'g', 'edge': 'k', 'scale': 1e1,
                                 'external_GMM': False}

        """
        self.his_.plot_histogram(
            bins=his_setting["bins"],
            upper_limitation=his_setting["upper_limitation"],
            lower_limitation=his_setting["lower_limitation"],
            step_range=his_setting["step_range"],
            face=his_setting["face"],
            edge=his_setting["edge"],
            Flag_GMM_fit=his_setting["Flag_GMM_fit"],
            max_n_components=his_setting["max_n_components"],
            scale=his_setting["scale"],
            external_GMM=his_setting["external_GMM"],
        )

    def plot_hist_2Dfit(self, his_setting):
        """This method plots histograms for 2D Gaussian fitting contrast for
        black PSFs, white PSFs and all together.

        Parameters
        ----------
        his_setting: dic
            This dictionary defines a histogram plotting setting. In the
            following you can see the example for it:

               | his_setting = {'bins': None, 'lower_limitation': -7e-4,
                                'upper_limitation': 7e-4, 'Flag_GMM_fit': True,
                                'max_n_components': 3, 'step_range': 1e-6,
                                'face': 'g', 'edge': 'k', }

        """
        self.his_.plot_fit_histogram(
            bins=his_setting["bins"],
            upper_limitation=his_setting["upper_limitation"],
            lower_limitation=his_setting["lower_limitation"],
            step_range=his_setting["step_range"],
            face=his_setting["face"],
            edge=his_setting["edge"],
            Flag_GMM_fit=his_setting["Flag_GMM_fit"],
            max_n_components=his_setting["max_n_components"],
        )

    def save_hist_data(self, dirName, name_dir, his_setting):
        """This function save the histogram data with HDF5 format.

        Parameters
        ----------
        dirName: str
            Path for saving data.

        name_dir: str
            Name that use for saving data.

        his_setting: dic
            This dictionary defines a histogram plotting setting. In the
            following you can see the example for it:

               | his_setting = {'lower_limitation': -7e-4, 'upper_limitation':
                                7e-4, 'Flag_GMM_fit': True, 'max_n_components':
                                3}

        """
        self.his_.save_hist_data(
            dirName=dirName,
            name=name_dir,
            upper_limitation=his_setting["upper_limitation"],
            lower_limitation=his_setting["lower_limitation"],
            Flag_GMM_fit=his_setting["Flag_GMM_fit"],
            max_n_components=his_setting["max_n_components"],
        )
        df_num = pd.DataFrame.from_dict(self.dic_video_num_particle)
        read_write_data.save_df2csv(df=df_num, path=dirName, name="list_video_num_particles")



from PySide6 import QtCore, QtWidgets
from PySide6.QtGui import Qt
from PySide6.QtWidgets import QLabel, QSlider


class Contrast_adjustment_GUI(QtWidgets.QMainWindow):
    min_intensity_signal = QtCore.Signal(object)
    max_intensity_signal = QtCore.Signal(object)
    alpha_signal = QtCore.Signal(object)
    beta_signal = QtCore.Signal(object)

    def __init__(self, alpha, beta, min_intensity, max_intensity, parent=None):
        super().__init__(parent)

        self.min_intensity = min_intensity
        self.max_intensity = max_intensity
        self.alpha = alpha
        self.beta = beta

        self.window = QtWidgets.QWidget()

        # Min hist
        self.slider_min_label = QLabel("Minimum:")
        self.slider_min = QSlider(QtWidgets.Horizontal)
        self.slider_min.setRange(min_intensity, max_intensity)
        self.slider_min.setValue(min_intensity)
        self.slider_min.setTracking(True)
        # self.slider_min.setTickPosition(QSlider.TicksBothSides)
        self.slider_min.valueChanged.connect(self.apply)

        # Max hist
        self.slider_max_label = QLabel("Maximum:")
        self.slider_max = QSlider(QtWidgets.Horizontal)
        self.slider_max.setRange(min_intensity, max_intensity)
        self.slider_max.setValue(max_intensity)
        self.slider_max.setTracking(True)
        # self.slider_max.setTickPosition(QSlider.TicksBothSides)
        self.slider_max.valueChanged.connect(self.apply)

        # Brightness
        self.slider_alpha_label = QLabel("Brightness:")
        self.slider_alpha = QSlider(QtWidgets.Horizontal)
        self.slider_alpha.setRange(1.0, 10.0)
        self.slider_alpha.setValue(alpha)
        self.slider_alpha.setTracking(True)
        # self.slider_alpha.setTickPosition(QSlider.TicksBothSides)
        self.slider_alpha.valueChanged.connect(self.apply)

        # Contrast
        self.slider_beta_label = QLabel("Contrast:")
        self.slider_beta = QSlider(QtWidgets.Horizontal)
        self.slider_beta.setRange(0, 100)
        self.slider_beta.setValue(beta)
        self.slider_beta.setTracking(True)
        # self.slider_beta.setTickPosition(QSlider.TicksBothSides)
        self.slider_beta.valueChanged.connect(self.apply)

        self.reset_btn = QtWidgets.QPushButton("Reset")
        self.reset_btn.setAutoDefault(False)
        self.reset_btn.clicked.connect(self.reset)
        self.reset_btn.setFixedWidth(100)

        self.auto_btn = QtWidgets.QPushButton("Auto")
        self.auto_btn.setAutoDefault(False)
        self.auto_btn.clicked.connect(self.auto)
        self.auto_btn.setFixedWidth(100)

        self.grid = QtWidgets.QGridLayout()
        self.grid.addWidget(self.createFirstExclusiveGroup(), 0, 0)
        self.grid.addWidget(self.auto_btn, 1, 0)
        self.grid.addWidget(self.reset_btn, 2, 0)

        self.setWindowTitle("Contrast_adjustment")
        self.window.setLayout(self.grid)
        self.window.show()

    def createFirstExclusiveGroup(self):
        self.groupBox_slideBar = QtWidgets.QGroupBox("Adjustment:")

        self.gridSlideBar = QtWidgets.QGridLayout()
        self.gridSlideBar.addWidget(self.slider_min_label, 0, 0)
        self.gridSlideBar.addWidget(self.slider_min, 0, 1)

        self.gridSlideBar.addWidget(self.slider_max_label, 1, 0)
        self.gridSlideBar.addWidget(self.slider_max, 1, 1)

        self.gridSlideBar.addWidget(self.slider_alpha_label, 2, 0)
        self.gridSlideBar.addWidget(self.slider_alpha, 2, 1)

        self.gridSlideBar.addWidget(self.slider_beta_label, 3, 0)
        self.gridSlideBar.addWidget(self.slider_beta, 3, 1)

        self.groupBox_slideBar.setLayout(self.gridSlideBar)

        return self.groupBox_slideBar

    def valueHandler(self, value):
        scaledValue = float(value) / 100

    def frame_getter(self, image):
        self.image = image

    def apply(self):
        self.alpha_signal.emit(self.slider_alpha.value())
        self.beta_signal.emit(self.slider_beta.value())
        self.min_intensity_signal.emit(self.slider_min.value())
        self.max_intensity_signal.emit(self.slider_max.value())

    def auto(self):
        self.alpha_signal.emit(None)
        self.beta_signal.emit(None)
        self.min_intensity_signal.emit(None)
        self.max_intensity_signal.emit(None)

        self.slider_min.setValue(self.min_intensity)
        self.slider_max.setValue(self.max_intensity)
        self.slider_alpha.setValue(self.alpha)
        self.slider_beta.setValue(self.beta)

    def reset(self):
        self.slider_min.setValue(0)
        self.slider_max.setValue(self.max_intensity)
        self.slider_alpha.setValue(5)
        self.slider_beta.setValue(50)

    def closeEvent(self, **kwargs):
        QtCore.QCoreApplication.instance().quit()



from __future__ import print_function

import imageio
import matplotlib
import matplotlib.animation as animation
import matplotlib.cm as cm
import matplotlib.patches as patches
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
import numpy as np
import pylab as pl
from joblib import Parallel, delayed
from matplotlib.patches import Circle
from mpl_toolkits.axes_grid1 import make_axes_locatable
from PySide6 import QtCore
from PySide6.QtCore import QRunnable, Signal, Slot
from scipy.ndimage import median_filter
from tqdm import tqdm

from piscat.InputOutput.cpu_configurations import CPUConfigurations
from piscat.Preproccessing.normalization import Normalization


class SignalConnection(QtCore.QObject):
    updateProgress_ = QtCore.Signal(int)
    finished = Signal()


class DisplaySubplot:
    def __init__(
        self, list_videos, numRows, numColumns, step=0, median_filter_flag=False, color="gray"
    ):
        """This class shows several videos (with the same number of frames) at
        once.

        Parameters
        ----------
        list_videos: list of NDArray
            List of videos.

        numRows: int
           It defines number of rows in sub-display.

        numColumns: int
           It defines number of columns in sub-display.

        median_filter_flag: bool
            In case it defines as True, a median filter is applied with size 3
            to remove hot pixel effect.

        color: str
            It defines the colormap for visualization.

        step: int
            Stride between visualization frames.

        """

        self.color = color
        self.jump = step
        self.pressed_key = {}
        self.memory = 0
        self.video = list_videos
        self.median_filter_flag = median_filter_flag
        self.numRows = numRows
        self.numColumns = numColumns
        self.numVideos = len(list_videos)

        self.max_numberFrames = np.max([vid_.shape[0] for vid_ in list_videos])

        self.fig, self.axes = plt.subplots(numRows, numColumns)

        self.im = []
        self.cb = []
        self.tx = []
        for vid_, ax in zip(self.video, self.axes):
            self.div = make_axes_locatable(ax)
            self.cax = self.div.append_axes("right", "5%", "5%")
            self.cv0 = vid_[0, :, :]
            im_ = ax.imshow(self.cv0, origin="lower", cmap=self.color)
            self.im.append(im_)
            self.cb.append(
                self.fig.colorbar(im_, cax=self.cax, format=ticker.FuncFormatter(self.fmt))
            )
            self.tx.append(ax.set_title("Frame 0"))

        self.ani = animation.FuncAnimation(
            self.fig, self.animate, frames=np.arange(0, self.max_numberFrames), blit=False
        )
        plt.show()

    def animate(self, i_):
        print(i_)
        self.fig.canvas.mpl_connect("key_press_event", self.press)
        if self.pressed_key.get("key") == "q":
            pl.close("all")

        i_ = i_ + self.memory
        if i_ < self.max_numberFrames:
            for vid_, im, tx in zip(self.video, self.im, self.tx):
                if self.median_filter_flag:
                    arr = median_filter(vid_[int(i_), :, :], 3)
                else:
                    arr = median_filter(vid_[int(i_), :, :], 3)
                vmax = np.max(arr)
                vmin = np.min(arr)
                im.set_data(arr)
                im.set_clim(vmin, vmax)
                tx.set_text("Frame {0}".format(i_))

                self.memory = i_ + self.jump
        else:
            self.ani.event_source.stop()

    def fmt(self, x, pos):
        a, b = "{:.2e}".format(x).split("e")
        b = int(b)
        return r"${} \times 10^{{{}}}$".format(a, b)

    def press(self, event):
        print("press", event.key)
        if event.key == "q":
            # close the current figure
            plt.close(event.canvas.figure)
            self.pressed_key["key"] = event.key


class Display:
    def __init__(self, video, step=0, color="gray", time_delay=0, median_filter_flag=False):
        """This class display the video.

        Parameters
        ----------
        video:(NDArray)
            Input video.

        median_filter_flag: bool
            In case it defines as True, a median filter is applied with size 3
            to remove hot pixel effect.

        color: str
            It defines the colormap for visualization.

        time_delay: float
            Delay between frames in milliseconds.

        step: int
            Stride between visualization frames.

        """
        self.video = video
        self.median_filter_flag = median_filter_flag
        self.jump = step
        self.time_delay = time_delay
        self.memory = 0

        self.fig = plt.figure()
        self.ax = self.fig.add_subplot(111)
        self.pressed_key = {}

        self.div = make_axes_locatable(self.ax)
        self.cax = self.div.append_axes("right", "5%", "5%")

        self.cv0 = self.video[0, :, :]
        self.im = self.ax.imshow(
            self.cv0, origin="lower", cmap=color
        )  # Here make an AxesImage rather than contour
        self.cb = self.fig.colorbar(self.im, cax=self.cax, format=ticker.FuncFormatter(self.fmt))
        self.tx = self.ax.set_title("Frame 0")

        self.ani = animation.FuncAnimation(
            self.fig,
            self.animate,
            frames=self.video.shape[0],
            blit=False,
            interval=self.time_delay,
            repeat=False,
            cache_frame_data=False,
        )
        plt.show()

    def animate(self, i_):
        self.fig.canvas.mpl_connect("key_press_event", self.press)
        if self.pressed_key.get("key") == "q":
            pl.close("all")

        i_ = i_ + self.memory
        if i_ < self.video.shape[0]:
            arr = self.video[i_, :, :]
            vmax = np.max(arr)
            vmin = np.min(arr)
            if self.median_filter_flag:
                frame_v = median_filter(arr, 3)
            else:
                frame_v = arr
            self.im.set_data(frame_v)
            self.im.set_clim(vmin, vmax)
            self.tx.set_text("Frame {0}".format(i_))
            self.memory = i_ + self.jump
        else:
            self.ani.event_source.stop()

    def fmt(self, x, pos):
        a, b = "{:.2e}".format(x).split("e")
        b = int(b)
        return r"${} \times 10^{{{}}}$".format(a, b)

    def press(self, event):
        print("press", event.key)
        if event.key == "q":
            # close the current figure
            plt.close(event.canvas.figure)
            self.pressed_key["key"] = event.key


class DisplayPSFs_subplotLocalizationDisplay:
    def __init__(
        self,
        list_videos,
        list_df_PSFs,
        list_titles,
        numRows,
        numColumns,
        color="gray",
        median_filter_flag=False,
        imgSizex=5,
        imgSizey=5,
        time_delay=0.1,
    ):
        """This class shows several videos (with the same number of frames) at
        once while highlight localize PSFs.

        Parameters
        ----------
        list_videos: list of NDArray
            List of videos

        list_df_PSFs: list panda data_frame
            List Data Frames that contains the location of PSFs for each video.

        numRows: int
            It defines number of rows in sub-display

        numColumns: int
            It defines number of columns in sub-display

        list_titles: list str
            List of titles for each sub plot

        median_filter_flag: bool
            In case it defines as True, a median filter is applied with size 3
            to remove hot pixel effect.

        color: str
            It defines the colormap for visualization.

        imgSizex: int
            Image length size.

        imgSizey: int
            Image width size.

        time_delay: float
            Delay between frames in milliseconds.

        """

        self.list_video = list_videos
        self.list_df_PSFs = list_df_PSFs
        self.median_filter_flag = median_filter_flag
        self.memory = 0
        self.numRows = numRows
        self.numColumns = numColumns
        self.numVideos = len(list_videos)
        if list_titles is None:
            self.list_titles = [None for _ in range(self.numVideos)]
        else:
            self.list_titles = list_titles

        self.fig = plt.figure(figsize=(imgSizex, imgSizey))
        self.imgGrid_list = []
        for i_ in range(1, self.numRows * self.numColumns + 1):
            img_grid_ = self.fig.add_subplot(self.numRows, self.numColumns, i_)
            img_grid_.axis("off")
            self.imgGrid_list.append(img_grid_)

        self.fig.tight_layout()
        self.fig.subplots_adjust(wspace=0.18)
        self.pressed_key = {}

        self.div_0 = []
        self.cax_0 = []
        self.im_0 = []
        self.cb_0 = []
        self.tx_0 = []

        for ax_, vid_, df_PSFs, tit_ in zip(
            self.imgGrid_list, self.list_video, self.list_df_PSFs, self.list_titles
        ):
            div_ = make_axes_locatable(ax_)
            self.div_0.append(div_)

            cax_ = div_.append_axes("right", "5%", "5%")
            self.cax_0.append(cax_)

            im_ = ax_.imshow(vid_[0, :, :], origin="lower", cmap=color)
            self.im_0.append(im_)
            self.cb_0.append(
                self.fig.colorbar(im_, cax=cax_, format=ticker.FuncFormatter(self.fmt))
            )
            if tit_ is not None:
                self.tx_0.append(ax_.set_title(tit_ + ", Frame 0"))
            else:
                self.tx_0.append(ax_.set_title("Frame 0"))

            particle = df_PSFs.loc[df_PSFs["frame"] == 0]
            particle_X = particle["x"].tolist()
            particle_Y = particle["y"].tolist()
            particle_sigma = particle["sigma"].tolist()

            for j_ in range(len(particle_X)):
                y = int(particle_Y[j_])
                x = int(particle_X[j_])
                sigma = particle_sigma[j_]
                ax_.add_patch(
                    Circle(
                        (x, y),
                        radius=np.sqrt(2) * sigma,
                        edgecolor="r",
                        facecolor="none",
                        linewidth=2,
                    )
                )

        self.ani = animation.FuncAnimation(
            self.fig,
            self.animate,
            frames=self.list_video[0].shape[0],
            blit=False,
            interval=time_delay,
            repeat=True,
            cache_frame_data=False,
        )
        plt.show()

    def animate(self, i_):
        self.fig.canvas.mpl_connect("key_press_event", self.press)
        if self.pressed_key.get("key") == "q":
            pl.close("all")
        if i_ < self.list_video[0].shape[0]:
            for idx_, (ax_, vid_, df_PSFs, tit_) in enumerate(
                zip(self.imgGrid_list, self.list_video, self.list_df_PSFs, self.list_titles)
            ):
                [p.remove() for p in reversed(ax_.patches)]

                arr = vid_[i_, :, :]
                vmax = np.max(arr)
                vmin = np.min(arr)
                if self.median_filter_flag:
                    frame_v = median_filter(arr, 3)
                else:
                    frame_v = arr

                self.im_0[idx_].set_data(frame_v)
                self.im_0[idx_].set_clim(vmin, vmax)

                if tit_ is not None:
                    self.tx_0[idx_].set_text(tit_ + ", Frame {}".format(i_))
                else:
                    self.tx_0[idx_].set_text("Frame {}".format(i_))

                particle = df_PSFs.loc[df_PSFs["frame"] == i_]
                particle_X = particle["x"].tolist()
                particle_Y = particle["y"].tolist()
                particle_sigma = particle["sigma"].tolist()
                for j_ in range(len(particle_X)):
                    y = int(particle_Y[j_])
                    x = int(particle_X[j_])
                    sigma = particle_sigma[j_]
                    ax_.add_patch(
                        Circle(
                            (x, y),
                            radius=np.sqrt(2) * sigma,
                            edgecolor="r",
                            facecolor="none",
                            linewidth=2,
                        )
                    )
        else:
            self.ani.event_source.stop()

    def fmt(self, x, pos):
        a, b = "{:.2e}".format(x).split("e")
        b = int(b)
        return r"${} \times 10^{{{}}}$".format(a, b)

    def press(self, event):
        print("press", event.key)
        if event.key == "q":
            # close the current figure
            plt.close(event.canvas.figure)
            self.pressed_key["key"] = event.key


class DisplayDataFramePSFsLocalization(QRunnable):
    def __init__(self, video, df_PSFs, time_delay=0.1, GUI_progressbar=False, *args, **kwargs):
        """
        This class displays video while highlighting PSFs.

        Parameters
        ----------
        video: NDArray
            Input video.

        df_PSFs: panda data_frame
            Data Frames that contains the location of PSFs.

        time_delay: float
            Delay between frames in milliseconds.

        GUI_progressbar: bool
            This actives GUI progress bar

        """
        self.cpu = CPUConfigurations()
        super(DisplayDataFramePSFsLocalization, self).__init__()

        self.video = video
        self.time_delay = time_delay
        self.df_PSFs = df_PSFs
        self.pressed_key = {}
        self.list_line = []
        if "particle" in self.df_PSFs.keys():
            self.list_particles_idx = self.df_PSFs.particle.unique()

        else:
            self.df_PSFs["particle"] = 0
            self.list_particles_idx = self.df_PSFs.particle.unique()

        self.GUI_progressbar = GUI_progressbar
        self.args = args
        self.kwargs = kwargs

        colors_ = cm.autumn(np.linspace(0, 1, len(self.list_particles_idx)))
        self.colors = colors_[0 : len(self.list_particles_idx), :]
        self.obj_connection = SignalConnection()

    @Slot()
    def run(self):
        self.gif_genrator(*self.args, **self.kwargs)
        if self.GUI_progressbar is True:
            self.obj_connection.finished.emit()

    def show_psf(
        self, jump=1, display_history=True, color_map="gray", save_flag=False, save_path=None
    ):
        self.norm_vid = Normalization(self.video).normalized_image_specific()
        img = None
        pl.ion()
        fig, ax = plt.subplots(1)
        ax.set_aspect("equal")
        if len(self.norm_vid.shape) == 3 and self.norm_vid.shape[0] > 0:
            for frame_number in range(0, self.norm_vid.shape[0] - jump, jump):
                if self.GUI_progressbar:
                    fig.canvas.mpl_connect("close_event", self.closeEvent)

                else:
                    fig.canvas.mpl_connect("key_press_event", self.press)

                if self.pressed_key.get("key") == "q":
                    pl.close("all")
                    break

                im = self.norm_vid[frame_number, :, :]
                pl.pause(self.time_delay)
                [p.remove() for p in reversed(ax.patches)]
                if img is None:
                    img = pl.imshow(im, cmap=color_map)
                    self.draw_circles(self.df_PSFs, ax, frame_number, color=self.colors)
                    if display_history:
                        [ln_.remove() for ln_ in self.list_line]
                        self.draw_trajectory(self.df_PSFs, ax, frame_number, color=self.colors)
                else:
                    img.set_data(im)
                    pl.draw()
                    pl.title("Press Q for exit\n Frame: " + str(frame_number))
                    self.draw_circles(self.df_PSFs, ax, frame_number, color=self.colors)
                    if display_history:
                        [ln_.remove() for ln_ in self.list_line]
                        self.draw_trajectory(self.df_PSFs, ax, frame_number, color=self.colors)
                    if save_flag:
                        pl.savefig(save_path + "fig" + str(frame_number) + ".png")

        else:
            im = self.norm_vid
            pl.pause(self.time_delay)
            [p.remove() for p in reversed(ax.patches)]
            if img is None:
                img = pl.imshow(im)
                self.draw_circles(self.df_PSFs, ax, frame_number=0, color=self.colors)
                if display_history:
                    [ln_.remove() for ln_ in self.list_line]
                    self.draw_trajectory(self.df_PSFs, ax, frame_number=0, color=self.colors)
            else:
                img.set_data(im)
                pl.draw()
                self.draw_circles(self.df_PSFs, ax, frame_number=0, color=self.colors)
                if display_history:
                    [ln_.remove() for ln_ in self.list_line]
                    self.draw_trajectory(self.df_PSFs, ax, frame_number=0, color=self.colors)

    def draw_circles(self, list_psf, ax, frame_number, color="red"):
        particle = list_psf.loc[list_psf["frame"] == frame_number]
        particle_X = particle["x"].tolist()
        particle_Y = particle["y"].tolist()
        particle_sigma = particle["sigma"].tolist()
        particle_labels = particle["particle"].tolist()
        for j_, p_label in zip(range(len(particle_X)), particle_labels):
            y = particle_Y[j_]
            x = particle_X[j_]
            sigma = particle_sigma[j_]
            color_idx = list(self.list_particles_idx).index(p_label)
            c = patches.Circle(
                (x, y),
                np.sqrt(2) * sigma,
                color=self.colors[color_idx],
                alpha=1,
                fill=False,
                linewidth=2,
            )

            ax.add_patch(c)
            pl.draw()

    def draw_trajectory(self, list_psf, ax, frame_number, color="red"):
        particle = list_psf.loc[list_psf["frame"] == frame_number]
        particle_labels = particle["particle"].tolist()

        self.list_line = []

        for label in particle_labels:
            all_particle_ = list_psf.loc[list_psf["particle"] == label]
            particle_f = all_particle_["frame"].tolist()
            particle_X = all_particle_["x"].tolist()
            particle_Y = all_particle_["y"].tolist()
            particle_sigma = all_particle_["sigma"].tolist()
            list_center = None
            for f, x, y, _ in zip(particle_f, particle_X, particle_Y, particle_sigma):
                if list_center is None:
                    center_position = [[x, y]]
                    list_center = center_position
                    list_center_array = np.asarray(list_center)

                else:
                    center_position = [x, y]
                    list_center.append(center_position)
                    list_center_array = np.asarray(list_center)

                if f <= frame_number:
                    color_idx = list(self.list_particles_idx).index(label)
                    (ln,) = ax.plot(
                        list_center_array[:, 0],
                        list_center_array[:, 1],
                        color=self.colors[color_idx],
                        linewidth=1,
                    )

                    self.list_line.append(ln)
                    pl.draw()

    def draw_trajectory_1(self, list_psf, ax, frame_number, color="red"):
        particle = list_psf.loc[list_psf["frame"] == frame_number]
        particle_labels = particle["particle"].tolist()
        list_line = []

        for label in particle_labels:
            all_particle_ = list_psf.loc[list_psf["particle"] == label]
            particle_f = all_particle_["frame"].tolist()
            particle_X = all_particle_["x"].tolist()
            particle_Y = all_particle_["y"].tolist()
            particle_sigma = all_particle_["sigma"].tolist()
            list_center = None
            flag_loop = True
            (ln,) = ax.plot(
                particle_X, particle_Y, color=self.colors[label], linewidth=1, alpha=0.7
            )
            list_line.append(ln)
        pl.draw()
        plt.pause(self.time_delay)
        [ln_.remove() for ln_ in list_line]

    def draw_trajectory_2(self, list_psf, ax, frame_number, color="red"):
        particle = list_psf.loc[list_psf["frame"] == frame_number]
        particle_labels = particle["particle"].tolist()
        list_line = []

        for label in particle_labels:
            all_particle_ = list_psf.loc[list_psf["particle"] == label]
            particle_f = all_particle_["frame"].tolist()
            particle_X = all_particle_["x"].tolist()
            particle_Y = all_particle_["y"].tolist()
            particle_sigma = all_particle_["sigma"].tolist()
            list_center = None
            flag_loop = True

            while flag_loop:
                list_line = []
                for f, x, y, _ in zip(particle_f, particle_X, particle_Y, particle_sigma):
                    if list_center is None:
                        center_position = [[x, y]]
                        list_center = center_position
                        list_center_array = np.asarray(list_center)
                    else:
                        center_position = [x, y]
                        list_center.append(center_position)
                        list_center_array = np.asarray(list_center)
                    if f <= frame_number:
                        (ln,) = ax.plot(
                            list_center_array[:, 0],
                            list_center_array[:, 1],
                            color=self.colors[label],
                            linewidth=1,
                            alpha=0.55,
                        )
                        list_line.append(ln)
                        pl.draw()
                        plt.pause(self.time_delay)
                    else:
                        flag_loop = False

                pl.draw()
                plt.pause(self.time_delay)
                [ln_.remove() for ln_ in list_line]

    def press(self, event):
        print("press", event.key)
        if event.key == "q":
            plt.close(event.canvas.figure)
            self.pressed_key["key"] = event.key

    def closeEvent(self, event):
        plt.close(event.canvas.figure)
        self.pressed_key["key"] = "q"

    def make_gif(self, frame_number, display_history=True, color_map="gray"):
        matplotlib.use("Agg")
        img = None
        pl.ioff()
        plt.ioff()

        fig, ax = plt.subplots(1)
        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)
        ax.set_aspect("equal")
        if len(self.norm_vid.shape) == 3 and self.norm_vid.shape[0] > 0:
            im = self.norm_vid[frame_number, :, :]
            pl.pause(self.time_delay)
            [p.remove() for p in reversed(ax.patches)]
            if img is None:
                img = pl.imshow(im, cmap=color_map)
                self.draw_circles(self.df_PSFs, ax, frame_number, color=self.colors)
                if display_history:
                    [ln_.remove() for ln_ in self.list_line]
                    self.draw_trajectory(self.df_PSFs, ax, frame_number, color=self.colors)
            else:
                img.set_data(im)
                pl.draw()
                pl.title(str(frame_number))
                self.draw_circles(self.df_PSFs, ax, frame_number, color=self.colors)
                if display_history:
                    [ln_.remove() for ln_ in self.list_line]
                    self.draw_trajectory(self.df_PSFs, ax, frame_number, color=self.colors)

            fig.canvas.draw()  # draw the canvas, cache the renderer
            image = np.frombuffer(fig.canvas.tostring_rgb(), dtype="uint8")
            image = image.reshape(fig.canvas.get_width_height()[::-1] + (3,))
            pl.close(fig)

            return image

    def gif_genrator(self, save_path, jump=5, fps=1.0):
        self.norm_vid = Normalization(self.video).normalized_image_specific()
        image_ = []
        if self.cpu.parallel_active is True:
            image_ = Parallel(
                n_jobs=self.cpu.n_jobs, backend=self.cpu.backend, verbose=self.cpu.verbose
            )(
                delayed(self.make_gif)(frame_number)
                for frame_number in range(0, self.norm_vid.shape[0] - jump, jump)
            )
            imageio.mimsave(save_path, image_, fps=fps)

        else:
            for frame_number in tqdm(range(0, self.norm_vid.shape[0] - jump, jump)):
                image_.append(self.make_gif(frame_number))
                if self.GUI_progressbar:
                    self.obj_connection.updateProgress_.emit(frame_number)
            imageio.mimsave(save_path, image_, fps=fps)


def histogram_of_each_frames(
    frame, bins="auto", range=None, normed=None, weights=None, density=None
):
    return np.histogram(
        frame, bins=bins, range=range, normed=normed, weights=weights, density=density
    )


def histogram_1D_signal(signal, bins="auto"):
    if bins == "auto":
        plt.hist(signal, bins=bins)  # arguments are passed to np.histogram
        plt.title("Histogram with 'auto' bins")
        plt.show()
    else:
        plt.hist(signal, bins=bins)  # arguments are passed to np.histogram
        plt.title("Histogram with" + str(bins) + "bins")
        plt.show()



import cmath


def L_theta(L, theta, p, ci, z, z_p):
    ni2sin2theta = p.ni**2 * cmath.sin(theta) ** 2
    sroot = cmath.sqrt(p.ns**2 - ni2sin2theta)
    groot = cmath.sqrt(p.ng**2 - ni2sin2theta)
    g0root = cmath.sqrt(p.ng0**2 - ni2sin2theta)
    i0root = cmath.sqrt(p.ni0**2 - ni2sin2theta)
    L[0] = (
        p.ni * (ci - z) * cmath.cos(theta)
        + z_p * sroot
        + p.tg * groot
        - p.tg0 * g0root
        - p.ti0 * i0root
    )
    L[1] = (
        p.ni
        * cmath.sin(theta)
        * (
            z
            - ci
            + p.ni
            * cmath.cos(theta)
            * (p.tg0 / g0root + p.ti0 / i0root - p.tg / groot - z_p / sroot)
        )
    )

    return L



from __future__ import print_function

import json
import os

import pandas as pd


class CPUConfigurations:
    def __init__(
        self,
        n_jobs=-1,
        backend="multiprocessing",
        verbose=0,
        parallel_active=True,
        threshold_for_parallel_run=None,
        flag_report=False,
    ):
        """This class generates a JSON file based on the parallel loop setting
        on the CPU that the user prefers.  This JSON was used by other
        functions and methods to set hyperparameters in a parallel loop.  For
        parallelization, PiSCAT used Joblib.

        | [1]. https://joblib.readthedocs.io/en/latest/generated/joblib.Parallel.html.

        Parameters
        ----------
        n_jobs: int
            The maximum number of workers that can work at the same time.
            If -1, all CPU cores are available for use.

        backend: str
            Specify the implementation of the parallelization backend.
            The following backends are supported:

            * `“loky”`:
                It can induce some communication and Memory overhead when
                exchanging input and output data with the worker Python
                processes.

            * `“multiprocessing”`:
                It previous process-based backend based on
                multiprocessing.Pool. Less robust than loky.

            * `“threading”`:
                It is a very low-overhead backend but it suffers from the
                Python Global Interpreter.  Lock if the called function relies
                a lot on Python objects. “threading” is mostly useful when the
                execution bottleneck is a compiled extension that explicitly
                releases the GIL (for instance a Cython loop wrapped in a “with
                nogil” block or an expensive call to a library such as NumPy).

        verbose: int, optional
            The verbosity level, if non zero, progress messages are
            printed. Above 50, the output is sent to stdout.  The frequency of
            the messages increases with the verbosity level. If it more than
            10, all iterations are reported.

        parallel_active: bool
            Functions will run the parallel implementation if it is True.

        threshold_for_parallel_run: float
            It reserved for next generation of PiSCAT.

        flag_report: bool
            This flag is set if you need to see the values that will be used
            for CPU configuration.

        """
        try:
            self.read_cpu_setting(flag_report)

        except FileNotFoundError:
            self.n_jobs = n_jobs
            self.backend = backend
            self.verbose = verbose
            self.parallel_active = parallel_active
            self.threshold_for_parallel_run = threshold_for_parallel_run

            setting_dic = {
                "n_jobs": [self.n_jobs],
                "backend": [self.backend],
                "verbose": [self.verbose],
                "parallel_active": [self.parallel_active],
                "threshold_for_parallel_run": [self.threshold_for_parallel_run],
            }

            self.save_cpu_setting(setting_dic)

    def save_cpu_setting(self, setting_dic):
        name = "cpu_configurations.json"
        here = os.path.dirname(os.getcwd())
        subdir = "piscat_configuration"

        try:
            dr_mk = os.path.join(here, subdir)
            os.mkdir(dr_mk)
            print("Directory ", subdir, " Created ")
        except FileExistsError:
            print("Directory ", subdir, " already exists")

        filepath = os.path.join(here, subdir, name)
        df_configfile = pd.DataFrame(data=setting_dic)
        df_configfile.to_json(filepath)

    def read_cpu_setting(self, flag_report=False):
        """
        flag_report: bool
               Whether you need to see the values that will be used for CPU configuration.
        """
        subdir = "piscat_configuration"
        here = os.path.dirname(os.getcwd())
        filepath = os.path.join(here, subdir, "cpu_configurations.json")

        with open(filepath) as json_file:
            cpu_setting = json.load(json_file)

        self.n_jobs = cpu_setting["n_jobs"]["0"]
        self.backend = cpu_setting["backend"]["0"]
        self.verbose = cpu_setting["verbose"]["0"]
        self.parallel_active = cpu_setting["parallel_active"]["0"]
        self.threshold_for_parallel_run = cpu_setting["threshold_for_parallel_run"]["0"]

        if flag_report:
            print("PiSCAT's general parallel flag is set to {}".format(self.parallel_active))
            print("\nThe number of parallel jobs is set to {}".format(self.n_jobs))
            print("\nThe backend is set to {}".format(self.backend))
            print("\nThe verbose is set to {}".format(self.verbose))



from .analysis_protein_histogram import ReadProteinAnalysis
from .analysis_protein_videos import protein_analysis
from .plot_protein_histogram import PlotProteinHistogram

__all__ = [
    "ReadProteinAnalysis",
    "protein_analysis",
    "PlotProteinHistogram",
]



from __future__ import print_function

import json
import os

import pandas as pd


class CameraParameters:
    def __init__(
        self,
        name,
        quantum_efficiency=0.3,
        electron_well_depth=180e3,
        max_electron_well_depth=200e3,
        bit_depth=12,
        pixelSize=0.66,
    ):
        """Based on the camera features, this class generates a JSON file.
        This JSON was used by other functions and methods to set certain
        parameters, such as pixel size.

        Parameters
        ----------
        name: str
            Name of camera.

        quantum_efficiency: float
            Quantum efficiency of camera.

        electron_well_depth: float
             Electron well depth of camera.

        max_electron_well_depth: float
            Maximum electron well depth of camera

        bit_depth: float
            Bit depth of the camera.

        pixelSize: float
            Pixel size of camera.

        """
        try:
            self.read_camera_setting(name=name)
        except FileNotFoundError:
            self.quantum_efficiency = quantum_efficiency
            self.electron_well_depth = electron_well_depth
            self.max_electron_well_depth = max_electron_well_depth
            self.bit_depth = bit_depth
            self.pixelSize = pixelSize

            self.setting_dic = {
                "quantum_efficiency": [self.quantum_efficiency],
                "electron_well_depth": [self.electron_well_depth],
                "max_electron_well_depth": [self.max_electron_well_depth],
                "bit_depth": [self.bit_depth],
                "pixelSize": [self.pixelSize],
            }
            self.save_camera_setting(name)

    def save_camera_setting(self, name="Photonfocus.json"):
        here = os.path.dirname(os.getcwd())
        subdir = "piscat_configuration"

        try:
            dr_mk = os.path.join(here, subdir)
            os.mkdir(dr_mk)
            print("Directory ", subdir, " Created ")
        except FileExistsError:
            print("Directory ", subdir, " already exists")

        filepath = os.path.join(here, subdir, name)
        df_configfile = pd.DataFrame(data=self.setting_dic)
        df_configfile.to_json(filepath)

    def read_camera_setting(self, name="Photonfocus.json"):
        subdir = "piscat_configuration"
        here = os.path.dirname(os.getcwd())
        filepath = os.path.join(here, subdir, name)

        with open(filepath) as json_file:
            cam_setting = json.load(json_file)

        self.quantum_efficiency = cam_setting["quantum_efficiency"]["0"]
        self.electron_well_depth = cam_setting["electron_well_depth"]["0"]
        self.max_electron_well_depth = cam_setting["max_electron_well_depth"]["0"]
        self.bit_depth = cam_setting["bit_depth"]["0"]
        self.pixelSize = cam_setting["pixelSize"]["0"]



from __future__ import print_function

import numpy as np
import pywt
from joblib import Parallel, delayed
from PySide6.QtCore import QObject, QRunnable, QThreadPool, Signal, Slot
from skimage import filters
from skimage.morphology import rectangle
from tqdm.notebook import tqdm

from piscat.InputOutput.cpu_configurations import CPUConfigurations
from piscat.Preproccessing.filtering import GuidedFilter
from piscat.Visualization.print_colors import PrintColors


class WorkerSignals(QObject):
    updateProgress_FPNC = Signal(int)
    result_FPNc = Signal(object)
    finished_FPNc = Signal()


class MedianProjectionFPNc(QRunnable):
    def __init__(self, video, select_correction_axis, flag_GUI=False):
        """This class uses a heuristic procedure called Median Projection FPN
        (mFPN) to reduce fixed pattern noise (FPN).

        References
        ----------
        [1] Mirzaalian Dastjerdi, Houman, et al. "Optimized analysis for
        sensitive detection and analysis of single proteins via interferometric
        scattering microscopy." Journal of Physics D: Applied Physics (2021).
        (http://iopscience.iop.org/article/10.1088/1361-6463/ac2f68)

        Parameters
        ----------
        video: NDArray
            The video is 3D-numpy (number of frames, width, height).

        select_correction_axis: int (0/1)
            * `0`: FPN will be applied row-wise.
            * `1`: FPN will be applied column-wise.

        """

        super(MedianProjectionFPNc, self).__init__()
        self.cpu = CPUConfigurations()
        self.video = video
        self.select_correction_axis = select_correction_axis
        self.flag_GUI = flag_GUI
        self.threadpool = QThreadPool()
        self.signals = WorkerSignals()

    @Slot()
    def run(self, *args, **kwargs):
        result = self.mFPNc(self.select_correction_axis)
        self.signals.result_FPNc.emit(result)

    def mFPNc(self, select_correction_axis):
        """
        Using the mPN approach on video.

        Parameters
        ----------
        select_correction_axis: int (0/1)

           * `0`: FPN will be applied row-wise.
           * `1`: FPN will be applied column-wise.

        Returns
        -------
        output: NDArray
            Video after using the mFPNc technique.
        """
        if False:
            print("\nFPN correction with parallel loop --->", end=" ")

            result = Parallel(
                n_jobs=self.cpu.n_jobs, backend="multiprocessing", verbose=self.cpu.verbose
            )(
                delayed(self.FPNc_kernel)(self.video[f_, :, :], select_correction_axis)
                for f_ in tqdm(range(0, self.video.shape[0]))
            )
            output = np.asarray(result)

        else:
            print("\nmedian FPN correction without parallel loop --->", end=" ")
            output = np.zeros_like(self.video)

            for i_ in tqdm(range(0, self.video.shape[0])):
                if self.flag_GUI is True:
                    self.signals.updateProgress_FPNC.emit(i_)

                diff = self.video[i_, :, :]

                gainMap1D_median = np.median(diff, axis=select_correction_axis)

                if select_correction_axis == 1:
                    FPN2D_median = np.repeat(
                        gainMap1D_median[:, np.newaxis],
                        diff.shape[select_correction_axis],
                        axis=select_correction_axis,
                    )
                elif select_correction_axis == 0:
                    FPN2D_median = np.repeat(
                        gainMap1D_median[np.newaxis, :],
                        diff.shape[select_correction_axis],
                        axis=select_correction_axis,
                    )

                output[i_, :, :] = diff - FPN2D_median

        if self.flag_GUI is True:
            self.signals.finished_FPNc.emit()
        print("Done")

        return output

    def FPNc_kernel(self, diff, select_correction_axis):
        gainMap1D_median = np.median(diff, axis=select_correction_axis)

        if select_correction_axis == 1:
            FPN2D_median = np.repeat(
                gainMap1D_median[:, np.newaxis],
                diff.shape[select_correction_axis],
                axis=select_correction_axis,
            )
        elif select_correction_axis == 0:
            FPN2D_median = np.repeat(
                gainMap1D_median[np.newaxis, :],
                diff.shape[select_correction_axis],
                axis=select_correction_axis,
            )

        AD_absDev = np.abs(diff - FPN2D_median)
        MAD_1D_MedianAD = np.median(AD_absDev, axis=select_correction_axis)

        if select_correction_axis == 1:
            MAD_2D = np.repeat(
                MAD_1D_MedianAD[:, np.newaxis],
                diff.shape[select_correction_axis],
                axis=select_correction_axis,
            )
        elif select_correction_axis == 0:
            MAD_2D = np.repeat(
                MAD_1D_MedianAD[np.newaxis, :],
                diff.shape[select_correction_axis],
                axis=select_correction_axis,
            )

        diff_f1f2_nan = diff.copy()
        diff_f1f2_nan[MAD_2D < AD_absDev] = np.nan
        gainMap1D_mean = np.nanmean(diff_f1f2_nan, axis=select_correction_axis)

        if select_correction_axis == 1:
            FPN2D = np.repeat(
                gainMap1D_mean[:, np.newaxis],
                diff.shape[select_correction_axis],
                axis=select_correction_axis,
            )
        elif select_correction_axis == 0:
            FPN2D = np.repeat(
                gainMap1D_mean[np.newaxis, :],
                diff.shape[select_correction_axis],
                axis=select_correction_axis,
            )

        return diff - FPN2D


class ColumnProjectionFPNc(QRunnable):
    def __init__(self, video, select_correction_axis, flag_GUI=False):
        """This class uses a heuristic procedure called Column Projection FPN
        (cpFPN) to reduce fixed pattern noise (FPN).

        Parameters
        ----------
        video: NDArray
            The video is 3D-numpy (number of frames, width, height).

        select_correction_axis: int (0/1)
            * `0`: FPN will be applied row-wise.
            * `1`: FPN will be applied column-wise.
        """
        super(ColumnProjectionFPNc, self).__init__()
        self.cpu = CPUConfigurations()
        self.video = video
        self.select_correction_axis = select_correction_axis
        self.flag_GUI = flag_GUI
        self.threadpool = QThreadPool()
        self.signals = WorkerSignals()

    @Slot()
    def run(self, *args, **kwargs):
        result = self.cpFPNc(self.select_correction_axis)
        self.signals.result_FPNc.emit(result)

    def cpFPNc(self, select_correction_axis):
        """
        Using the cpFPN approach on video.

        Parameters
        ----------
        select_correction_axis: int (0/1)

           * `0`: FPN will be applied row-wise.
           * `1`: FPN will be applied column-wise.

        Returns
        -------
        output: NDArray
            Video after using the cpFPNc technique.
        """
        if False:
            print("\nFPN correction with parallel loop --->", end=" ")

            result = Parallel(
                n_jobs=self.cpu.n_jobs, backend="multiprocessing", verbose=self.cpu.verbose
            )(
                delayed(self.FPNc_kernel)(self.video[f_, :, :], select_correction_axis)
                for f_ in tqdm(range(0, self.video.shape[0]))
            )
            output = np.asarray(result)

        else:
            print("\ncpFPN correction without parallel loop --->", end=" ")
            output = np.zeros_like(self.video)

            for i_ in tqdm(range(0, self.video.shape[0])):
                if self.flag_GUI is True:
                    self.signals.updateProgress_FPNC.emit(i_)

                diff = self.video[i_, :, :]

                gainMap1D_median = np.median(diff, axis=select_correction_axis)

                if select_correction_axis == 1:
                    FPN2D_median = np.repeat(
                        gainMap1D_median[:, np.newaxis],
                        diff.shape[select_correction_axis],
                        axis=select_correction_axis,
                    )
                elif select_correction_axis == 0:
                    FPN2D_median = np.repeat(
                        gainMap1D_median[np.newaxis, :],
                        diff.shape[select_correction_axis],
                        axis=select_correction_axis,
                    )

                AD_absDev = np.abs(diff - FPN2D_median)
                MAD_1D_MedianAD = np.median(
                    AD_absDev, axis=select_correction_axis
                )  # convert to median from here

                if select_correction_axis == 1:
                    MAD_2D = np.repeat(
                        MAD_1D_MedianAD[:, np.newaxis],
                        diff.shape[select_correction_axis],
                        axis=select_correction_axis,
                    )
                elif select_correction_axis == 0:
                    MAD_2D = np.repeat(
                        MAD_1D_MedianAD[np.newaxis, :],
                        diff.shape[select_correction_axis],
                        axis=select_correction_axis,
                    )

                diff_f1f2_nan = diff.copy()
                diff_f1f2_nan[MAD_2D < AD_absDev] = np.nan
                gainMap1D_mean = np.nanmean(diff_f1f2_nan, axis=select_correction_axis)

                if select_correction_axis == 1:
                    FPN2D = np.repeat(
                        gainMap1D_mean[:, np.newaxis],
                        diff.shape[select_correction_axis],
                        axis=select_correction_axis,
                    )
                elif select_correction_axis == 0:
                    FPN2D = np.repeat(
                        gainMap1D_mean[np.newaxis, :],
                        diff.shape[select_correction_axis],
                        axis=select_correction_axis,
                    )

                output[i_, :, :] = diff - FPN2D

        if self.flag_GUI is True:
            self.signals.finished_FPNc.emit()
        print("Done")

        return output

    def FPNc_kernel(self, diff, select_correction_axis):
        gainMap1D_median = np.median(diff, axis=select_correction_axis)

        if select_correction_axis == 1:
            FPN2D_median = np.repeat(
                gainMap1D_median[:, np.newaxis],
                diff.shape[select_correction_axis],
                axis=select_correction_axis,
            )
        elif select_correction_axis == 0:
            FPN2D_median = np.repeat(
                gainMap1D_median[np.newaxis, :],
                diff.shape[select_correction_axis],
                axis=select_correction_axis,
            )

        AD_absDev = np.abs(diff - FPN2D_median)
        MAD_1D_MedianAD = np.median(AD_absDev, axis=select_correction_axis)

        if select_correction_axis == 1:
            MAD_2D = np.repeat(
                MAD_1D_MedianAD[:, np.newaxis],
                diff.shape[select_correction_axis],
                axis=select_correction_axis,
            )
        elif select_correction_axis == 0:
            MAD_2D = np.repeat(
                MAD_1D_MedianAD[np.newaxis, :],
                diff.shape[select_correction_axis],
                axis=select_correction_axis,
            )

        diff_f1f2_nan = diff.copy()
        diff_f1f2_nan[MAD_2D < AD_absDev] = np.nan
        gainMap1D_mean = np.nanmean(diff_f1f2_nan, axis=select_correction_axis)

        if select_correction_axis == 1:
            FPN2D = np.repeat(
                gainMap1D_mean[:, np.newaxis],
                diff.shape[select_correction_axis],
                axis=select_correction_axis,
            )
        elif select_correction_axis == 0:
            FPN2D = np.repeat(
                gainMap1D_mean[np.newaxis, :],
                diff.shape[select_correction_axis],
                axis=select_correction_axis,
            )

        return diff - FPN2D


class FrequencyFPNc:
    def __init__(self, video, inter_flag_parallel_active=True):
        """This class corrects FPN using two well-known frequency domain
        techniques from the literature.

        References
        ----------
        [1] Cao, Yanlong, et al. "A multi-scale non-uniformity correction
        method based on wavelet decomposition and guided filtering for uncooled
        long wave infrared camera." Signal Processing: Image Communication 60
        (2018): 13-21.

        [2] Zeng, Qingjie, et al. "Single infrared image-based stripe
        non-uniformity correction via a two-stage filtering method." Sensors
        18.12 (2018): 4299.

        Parameters
        ----------
        video: NDArray
            The video is 3D-numpy (number of frames, width, height).

        inter_flag_parallel_active: bool
            If the user wants to enable generic parallel tasks in CPU
            configuration, this flag is used to disable parallel execution of
            this function.

        """
        super(FrequencyFPNc, self).__init__()
        self.cpu = CPUConfigurations()
        self.inter_flag_parallel_active = inter_flag_parallel_active

        self.coeffs = None
        self.arr = None
        self.coeff_slices = None
        self.video_out = None
        self.im_fft_noiseless = None
        self.im_fft_noise = None
        self.im_recon = None
        self.coeffs_result = None
        self.print_color = PrintColors()
        self.video = video

    def update_fFPN(self, direction="Horizontal", max_iterations=10, width=1):
        """
        This method corrects the FPNc by using FFT [2].

        Parameters
        ----------
        direction: str
            Axis that FPN correction should apply on it.

            * ``'Horizontal'``
            * ``'Vertical'``

        max_iterations: int
            Total number of filtering iterations.

        width: int
            The frequency mask's width.

        Returns
        -------
        n_video: NDArray
            Video after using the fFPNc technique.
        """
        self.max_iterations = max_iterations
        self.video_out = np.ndarray(self.video.shape)
        self.direction = direction

        w = width
        if self.direction == "Horizontal":
            self.mask = np.ones(self.video[0, :, :].shape)
            center_freq_line_H = round(1 + (0.5 * self.mask.shape[1]))
            self.mask[center_freq_line_H - w : center_freq_line_H + w, :] = 0

        elif self.direction == "Vertical":
            self.mask = np.ones(self.video[0, :, :].shape)
            center_freq_line_v = round(1 + (0.5 * self.mask.shape[1]))
            self.mask[:, center_freq_line_v - w : center_freq_line_v + w] = 0

        if self.cpu.parallel_active is True and self.inter_flag_parallel_active is True:
            print("\n---start fFPNc with Parallel---")
            result = Parallel(
                n_jobs=self.cpu.n_jobs, backend=self.cpu.backend, verbose=self.cpu.verbose
            )(delayed(self.fft_kernel)(f_) for f_ in tqdm(range(self.video.shape[0])))

            self.video_out = np.asarray(result)
        else:
            print("\n---start fFPNc without Parallel---")
            result = [self.fft_kernel(f_) for f_ in tqdm(range(0, self.video.shape[0]))]
            self.video_out = np.asarray(result)

        return self.video_out

    def fft_kernel(self, f_):
        selected_frame = self.video[f_, :, :]
        im_fft = np.fft.fftshift(np.fft.fft2(selected_frame))
        amplitude = np.abs(im_fft)
        phase = np.angle(im_fft)

        amplitude_noiseless = np.multiply(self.mask, amplitude)
        im_fft_noiseless = np.multiply(amplitude_noiseless, np.exp(phase * 1j))
        im_fft_noiseless = np.fft.fftshift(im_fft_noiseless)
        im_noiseless = np.fft.ifft2(im_fft_noiseless)

        im_noise = self.video[f_, :, :] - np.real(im_noiseless)

        if self.direction == "Horizontal":
            selem = rectangle(nrows=1, ncols=5)

        elif self.direction == "Vertical":
            selem = rectangle(nrows=5, ncols=1)

        filter_img = im_noise.copy()
        for _ in range(self.max_iterations):
            try:
                filter_img = filters.gaussian(filter_img, sigma=1.2, preserve_range=True)
                # filter_img = rank.mean(filter_img, selem=selem)

            except:
                print(f"{self.WARNING}\nThe Gaussian filter can not work!{self.ENDC}")

        img_FPNc = np.real(im_noiseless) + np.real(filter_img)

        return img_FPNc

    def update_wFPN(self, direction="Horizontal"):
        """
        This method corrects the FPNc by using wavelet[1].

        Parameters
        ----------
        direction: str
            Axis that FPN correction should apply on it.

            * ``'Horizontal'``
            * ``'Vertical'``

        Returns
        -------
        n_video: NDArray
            Video after using the wFPNc technique.
        """
        self.video_out = np.empty_like(self.video)

        if self.cpu.parallel_active is True and self.inter_flag_parallel_active is True:
            print("\n---start wFPNc with Parallel---")
            result = Parallel(
                n_jobs=self.cpu.n_jobs, backend=self.cpu.backend, verbose=self.cpu.verbose
            )(
                delayed(self.wavelet_denoise)(f_, direction)
                for f_ in tqdm(range(self.video.shape[0]))
            )

            self.video_out = np.asarray(result)
        else:
            print("\n---start wFPNc without Parallel---")
            for f_ in tqdm(range(0, self.video.shape[0])):
                self.video_out[f_, :, :] = self.wavelet_denoise(f_, direction=direction)

        return self.video_out

    def wavelet_denoise(self, f_, direction="Horizontal"):
        coeffs_1 = pywt.dwt2(self.video[f_, :, :], wavelet="sym8")
        cA_1, (cH_1, cV_1, cD_1) = coeffs_1

        coeffs_2 = pywt.dwt2(cA_1, wavelet="sym8")
        cA_2, (cH_2, cV_2, cD_2) = coeffs_2

        coeffs_3 = pywt.dwt2(cA_2, wavelet="sym8")
        cA_3, (cH_3, cV_3, cD_3) = coeffs_3

        if direction == "Horizontal":
            cH_3_corr = self.guided_filter(
                cH_3, cA_3, win_size=(1, np.int64(0.5 * cA_3.shape[0])), eps=0.3**2
            )
            coeffs_3_corr = cA_3, (cH_3_corr, cV_3, cD_3)
            cA_2_new_ = pywt.idwt2(coeffs_3_corr, "sym8")
            cA_2_new = self.size_handling(cH_2, cA_2_new_)

            cH_2_corr = self.guided_filter(
                cH_2, cA_2_new, win_size=(1, np.int64(0.25 * cA_3.shape[0])), eps=0.2**2
            )
            coeffs_2_corr = cA_2_new, (cH_2_corr, cV_2, cD_2)
            cA_1_new_ = pywt.idwt2(coeffs_2_corr, "sym8")
            cA_1_new = self.size_handling(cH_1, cA_1_new_)

            cH_1_corr = self.guided_filter(
                cH_1, cA_1_new, win_size=(1, np.int64(0.1 * cA_3.shape[0])), eps=0.1**2
            )
            coeffs_1_corr = cA_1_new, (cH_1_corr, cV_1, cD_1)
            im_corr = pywt.idwt2(coeffs_1_corr, "sym8")
            return self.size_handling(self.video[f_, :, :], im_corr)

        elif direction == "Vertical":
            cV_3_corr = self.guided_filter(
                cV_3, cA_3, win_size=(np.int64(0.5 * cA_3.shape[0]), 1), eps=0.3**2
            )
            coeffs_3_corr = cA_3, (cH_3, cV_3_corr, cD_3)
            cA_2_new_ = pywt.idwt2(coeffs_3_corr, "sym8")
            cA_2_new = self.size_handling(cV_2, cA_2_new_)

            cV_2_corr = self.guided_filter(
                cV_2, cA_2_new, win_size=(np.int64(0.25 * cA_3.shape[0]), 1), eps=0.2**2
            )
            coeffs_2_corr = cA_2_new, (cH_2, cV_2_corr, cD_2)
            cA_1_new_ = pywt.idwt2(coeffs_2_corr, "sym8")
            cA_1_new = self.size_handling(cV_1, cA_1_new_)

            cV_1_corr = self.guided_filter(
                cV_1, cA_1_new, win_size=(np.int64(0.1 * cA_3.shape[0]), 1), eps=0.1**2
            )
            coeffs_1_corr = cA_1_new, (cH_1, cV_1_corr, cD_1)
            im_corr = pywt.idwt2(coeffs_1_corr, "sym8")
            return self.size_handling(self.video[f_, :, :], im_corr)

    def guided_filter(self, input_image, guided_image, win_size, eps):
        GF_ = GuidedFilter(guided_image, radius=win_size, eps=eps)
        return GF_.filter(input_image)

    def size_handling(self, im_1, im_2):
        if im_1.shape[0] == im_2.shape[0] and im_1.shape[1] == im_2.shape[1]:
            im_out = im_2
        elif im_1.shape[0] == im_2.shape[0]:
            num_cut = np.abs(im_1.shape[1] - im_2.shape[1])
            im_out = im_2[:, 0:-num_cut]
        elif im_1.shape[1] == im_2.shape[1]:
            num_cut = np.abs(im_1.shape[0] - im_2.shape[0])
            im_out = im_2[0:-num_cut, :]
        else:
            num_cut_0 = np.abs(im_1.shape[0] - im_2.shape[0])
            num_cut_1 = np.abs(im_1.shape[1] - im_2.shape[1])
            im_out = im_2[0:-num_cut_0, 0:-num_cut_1]
        return im_out



from PySide6 import QtCore, QtWidgets


class Tracking_GUI(QtWidgets.QWidget):
    output_setting_Tab_tracking = QtCore.Signal(object)
    update_tab_index = QtCore.Signal(int)

    def __init__(self, parent=None):
        super(Tracking_GUI, self).__init__(parent)

        self.his_all_particles = None
        self.memory = None
        self.search_range = None
        self.temporal_length = None

        self.setting_tracking = {}

        self.empty_value_box_flag = False
        self.empty_value_optional_box_flag_1 = False
        self.empty_value_optional_box_flag_2 = False

        self.resize(300, 300)
        self.setWindowTitle("PSFs Tracking")

        self.btn_filterLinking = QtWidgets.QPushButton("Next")
        self.btn_filterLinking.setFixedWidth(120)
        self.btn_filterLinking.setFixedHeight(30)
        self.btn_filterLinking.clicked.connect(self.do_update_tfilter)

        self.line_memory = QtWidgets.QLineEdit()
        self.line_memory.setPlaceholderText("Memory")
        self.line_memory_label = QtWidgets.QLabel("Memory (frame):")

        self.line_search_range = QtWidgets.QLineEdit()
        self.line_search_range.setPlaceholderText("Move")
        self.line_searchRange_label = QtWidgets.QLabel("Neighborhood size(px):")

        self.line_min_V_shape = QtWidgets.QLineEdit()
        self.line_min_V_shape.setPlaceholderText("Minimum_temporal_length (frame)")
        self.line_min_V_shape_label = QtWidgets.QLabel("Minimum_temporal_length (frame):")

        self.grid = QtWidgets.QGridLayout()
        self.grid.addWidget(self.createFirstExclusiveGroup(), 0, 0)

        self.setLayout(self.grid)

    def __del__(self):
        print("Destructor called, Employee deleted.")

    def createFirstExclusiveGroup(self):
        groupBox = QtWidgets.QGroupBox("Linking")
        self.checkbox_sorting_based_lenght = QtWidgets.QCheckBox("Sorting", self)

        self.grid_linking = QtWidgets.QGridLayout()

        self.grid_linking.addWidget(self.line_memory_label, 0, 0)
        self.grid_linking.addWidget(self.line_memory, 0, 1)

        self.grid_linking.addWidget(self.line_searchRange_label, 1, 0)
        self.grid_linking.addWidget(self.line_search_range, 1, 1)

        self.grid_linking.addWidget(self.line_min_V_shape_label, 2, 0)
        self.grid_linking.addWidget(self.line_min_V_shape, 2, 1)
        self.grid_linking.addWidget(self.btn_filterLinking, 3, 0)

        self.grid_linking.addWidget(self.checkbox_sorting_based_lenght, 3, 3)

        groupBox.setLayout(self.grid_linking)
        return groupBox

    def do_update_tfilter(self):
        self.get_values_1()
        self.get_values_2()
        if self.empty_value_box_flag:
            self.setting_tracking["Memory (frame)"] = self.memory
            self.setting_tracking["Neighborhood_size (px)"] = self.search_range
            self.setting_tracking["Minimum_temporal_length (frame)"] = self.temporal_length
            self.output_setting_Tab_tracking.emit(self.setting_tracking)
            self.update_tab_index.emit(4)

    def get_values_1(self):
        try:
            self.memory = int(self.line_memory.text())
            self.search_range = int(self.line_search_range.text())

            self.empty_value_box_flag = True

        except:
            self.msg_box3 = QtWidgets.QMessageBox()
            self.msg_box3.setWindowTitle("Warning!")
            self.msg_box3.setText("Please filled all parameters!")
            self.msg_box3.exec_()

            self.empty_value_box_flag = False

    def get_values_2(self):
        try:
            self.temporal_length = int(self.line_min_V_shape.text())

            self.empty_value_box_flag = True

        except:
            self.msg_box3 = QtWidgets.QMessageBox()
            self.msg_box3.setWindowTitle("Warning!")
            self.msg_box3.setText("Please filled all parameters!")
            self.msg_box3.exec_()

            self.empty_value_box_flag = False

    def closeEvent(self, event):
        QtCore.QCoreApplication.instance().quit()



import math

import matplotlib.pyplot as plt
import numpy as np
import scipy as sp


class DirectionalIntensity:
    def __init__(self):
        pass

    def interpolate_pixels_along_line(self, x0, y0, x1, y1):
        """Uses Xiaolin Wu's line algorithm to interpolate all of the pixels along a
        straight line, given two points (x0, y0) and (x1, y1)

        References
        ----------
        Adapted from:
        [1] https://stackoverflow.com/questions/3798333/image-information-along-a-polar-coordinate-system
        [2] Wikipedia article containing pseudo code that function was based
        off of: http://en.wikipedia.org/wiki/Xiaolin_Wu's_line_algorithm
        """
        pixels = []
        steep = abs(y1 - y0) > abs(x1 - x0)

        # Ensure that the path to be interpolated is shallow and from left to right
        if steep:
            t = x0
            x0 = y0
            y0 = t

            t = x1
            x1 = y1
            y1 = t

        if x0 > x1:
            t = x0
            x0 = x1
            x1 = t

            t = y0
            y0 = y1
            y1 = t

        dx = x1 - x0
        dy = y1 - y0
        gradient = dy / dx  # slope

        # Get the first given coordinate and add it to the return list
        x_end = round(x0)
        y_end = y0 + (gradient * (x_end - x0))
        xpxl0 = x_end
        ypxl0 = round(y_end)
        if steep:
            pixels.extend([(ypxl0, xpxl0), (ypxl0 + 1, xpxl0)])
        else:
            pixels.extend([(xpxl0, ypxl0), (xpxl0, ypxl0 + 1)])

        interpolated_y = y_end + gradient

        # Get the second given coordinate to give the main loop a range
        x_end = round(x1)
        y_end = y1 + (gradient * (x_end - x1))
        xpxl1 = x_end
        ypxl1 = round(y_end)

        # Loop between the first x coordinate and the second x coordinate,
        # interpolating the y coordinates
        for x in range(xpxl0 + 1, xpxl1):
            if steep:
                pixels.extend(
                    [(math.floor(interpolated_y), x), (math.floor(interpolated_y) + 1, x)]
                )

            else:
                pixels.extend(
                    [(x, math.floor(interpolated_y)), (x, math.floor(interpolated_y) + 1)]
                )

            interpolated_y += gradient

        # Add the second given coordinate to the given list
        if steep:
            pixels.extend([(ypxl1, xpxl1), (ypxl1 + 1, xpxl1)])
        else:
            pixels.extend([(xpxl1, ypxl1), (xpxl1, ypxl1 + 1)])

        return pixels

    def radial_profile_app2(self, data, r):
        R = data.shape[0] // 2  # Radial profile radius
        range_arr = np.arange(-R, R + 1)
        ids = (range_arr[:, None] ** 2 + range_arr**2).ravel()
        count = np.bincount(ids)

        R0 = R + 1
        dists = np.unique(r[:R0, :R0][np.tril(np.ones((R0, R0), dtype=bool))])

        mean_data = (np.bincount(ids, data.ravel()) / count)[count != 0]
        return dists, mean_data

    def plot_directional_intensity(self, data, origin=None):
        """
        Makes a cicular histogram showing average intensity binned by direction
        from "origin" for each band in "data" (a 3D numpy array). "origin" defaults
        to the center of the image.
        """

        def intensity_rose(theta, band, color):
            theta, band = theta.flatten(), band.flatten()
            intensities, theta_bins = self.bin_by(band, theta)
            mean_intensity = map(np.mean, intensities)
            width = np.diff(theta_bins)[0]
            plt.bar(theta_bins, mean_intensity, width=width, color=color)
            plt.xlabel(color + " Band")
            plt.yticks([])

        # Make cartesian coordinates for the pixel indicies
        # (The origin defaults to the center of the image)
        x, y = self.index_coords(data, origin)

        # Convert the pixel indices into polar coords.
        r, theta = self.cart2polar(x, y)

        # Unpack bands of the image
        # red, green, blue = data.T
        red = data.T
        green = data.T
        blue = data.T
        # Plot...
        plt.figure()

        plt.subplot(2, 2, 1, projection="polar")
        intensity_rose(theta, red, "Red")

        plt.subplot(2, 2, 2, projection="polar")
        intensity_rose(theta, green, "Green")

        plt.subplot(2, 1, 2, projection="polar")
        intensity_rose(theta, blue, "Blue")

        plt.suptitle("Average intensity as a function of direction")

    def plot_polar_image(self, data, origin=None):
        """Plots an image reprojected into polar coordinages with the origin
        at "origin" (a tuple of (x0, y0), defaults to the center of the image)"""
        polar_grid, r, theta = self.reproject_image_into_polar(data, origin)
        plt.figure()
        plt.imshow(polar_grid, extent=(theta.min(), theta.max(), r.max(), r.min()))
        plt.axis("auto")
        plt.ylim(plt.ylim()[::-1])
        plt.xlabel("Theta Coordinate (radians)")
        plt.ylabel("R Coordinate (pixels)")
        plt.title("Image in Polar Coordinates")

    def index_coords(self, data, origin=None):
        """Creates x & y coords for the indicies in a numpy array "data".
        "origin" defaults to the center of the image. Specify origin=(0,0)
        to set the origin to the lower left corner of the image."""
        ny, nx = data.shape[:2]
        if origin is None:
            origin_x, origin_y = nx // 2, ny // 2
        else:
            origin_x, origin_y = origin
        x, y = np.meshgrid(np.arange(nx), np.arange(ny))
        x -= origin_x
        y -= origin_y
        return x, y

    def cart2polar(self, x, y):
        r = np.sqrt(x**2 + y**2)
        theta = np.arctan2(y, x)
        return r, theta

    def polar2cart(self, r, theta):
        x = r * np.cos(theta)
        y = r * np.sin(theta)
        return x, y

    def bin_by(self, x, y, nbins=30):
        """Bin x by y, given paired observations of x & y.
        Returns the binned "x" values and the left edges of the bins."""
        bins = np.linspace(y.min(), y.max(), nbins + 1)
        # To avoid extra bin for the max value
        bins[-1] += 1

        indicies = np.digitize(y, bins)

        output = []
        for i in range(1, len(bins)):
            output.append(x[indicies == i])

        # Just return the left edges of the bins
        bins = bins[:-1]

        return output, bins

    def reproject_image_into_polar(self, data, origin=None):
        """Reprojects a 3D numpy array ("data") into a polar coordinate system.
        "origin" is a tuple of (x0, y0) and defaults to the center of the image."""
        ny, nx = data.shape[:2]
        if origin is None:
            origin = (nx // 2, ny // 2)

        # Determine that the min and max r and theta coords will be...
        x, y = self.index_coords(data, origin=origin)
        r, theta = self.cart2polar(x, y)

        # Make a regular (in polar space) grid based on the min and max r & theta
        r_i = np.linspace(r.min(), r.max(), nx)
        theta_i = np.linspace(theta.min(), theta.max(), ny)
        theta_grid, r_grid = np.meshgrid(theta_i, r_i)

        # Project the r and theta grid back into pixel coordinates
        xi, yi = self.polar2cart(r_grid, theta_grid)
        xi += origin[0]  # We need to shift the origin back to
        yi += origin[1]  # back to the lower-left corner...
        xi, yi = xi.flatten(), yi.flatten()
        coords = np.vstack((xi, yi))  # (map_coordinates requires a 2xn array)

        # Reproject each band individually and the restack
        # (uses less Memory than reprojection the 3-dimensional array in one step)
        bands = []
        for band in data.T:
            zi = sp.ndimage.map_coordinates(band, coords, order=1)
            bands.append(zi.reshape((nx, ny)))
        output = np.dstack(bands)
        return output, r_i, theta_i



from .filtering import (
    FFT2D,
    FastRadialSymmetryTransform,
    Filters,
    GrayGuidedFilter,
    GuidedFilter,
    MultiDimGuidedFilter,
    RadialVarianceTransform,
)
from .FPNc import ColumnProjectionFPNc, FrequencyFPNc, MedianProjectionFPNc
from .normalization import Normalization
from .patch_genrator import ImagePatching

__all__ = [
    "FFT2D",
    "FastRadialSymmetryTransform",
    "Filters",
    "GrayGuidedFilter",
    "GuidedFilter",
    "MultiDimGuidedFilter",
    "RadialVarianceTransform",
    "ColumnProjectionFPNc",
    "FrequencyFPNc",
    "MedianProjectionFPNc",
    "Normalization",
    "ImagePatching",
]



import json
import os

import pandas as pd
from tensorflow.python.client import device_lib

from piscat.Visualization.print_colors import PrintColors

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"


class GPUConfigurations(PrintColors):
    def __init__(self, gpu_device=None, gpu_active_flag=True, flag_report=False):
        """This class generates a JSON file for setting on the GPU that the
        user prefers.  This JSON was used by other functions and methods to set
        hyperparameters in a gpu usage.  For parallelization, PiSCAT used
        Tensorflow.

        Parameters
        ----------
        gpu_device: int
            Select the GPU device that will be used.

        gpu_active_flag: bool
            Turn on the GPU version of the code. Otherwise, code is executed on
            the CPU.

        flag_report: bool
            This flag is set if you need to see the values that will be used
            for CPU configuration.

        """

        PrintColors.__init__(self)

        try:
            self.read_gpu_setting(flag_report)

        except FileNotFoundError:
            list_gpu_cpu = self.get_available_devices()
            count_ = 0
            for gpu_ in list_gpu_cpu:
                if "GPU" in gpu_[0]:
                    count_ = count_ + 1
                    print(str(gpu_))

            if count_ == 0:
                self.gpu_active_flag = False
                self.gpu_device = None
            else:
                self.gpu_active_flag = True
                if gpu_device is None:
                    self.gpu_device = 0
                else:
                    self.gpu_device = gpu_device

            setting_dic = {"gpu_active": [self.gpu_active_flag], "gpu_device": [self.gpu_device]}

            self.save_gpu_setting(setting_dic)

    def save_gpu_setting(self, setting_dic):
        name = "gpu_configurations.json"
        here = os.path.dirname(os.getcwd())
        subdir = "piscat_configuration"

        try:
            dr_mk = os.path.join(here, subdir)
            os.mkdir(dr_mk)
            print("Directory ", subdir, " Created ")
        except FileExistsError:
            print("Directory ", subdir, " already exists")

        filepath = os.path.join(here, subdir, name)
        df_configfile = pd.DataFrame(data=setting_dic)
        df_configfile.to_json(filepath)

    def read_gpu_setting(self, flag_report=False):
        """
        Parameters
        ----------
         flag_report: bool
            Whether you need to see the values that will be used for CPU configuration.

        """
        subdir = "piscat_configuration"
        here = os.path.dirname(os.getcwd())
        filepath = os.path.join(here, subdir, "gpu_configurations.json")

        with open(filepath) as json_file:
            gpu_setting = json.load(json_file)

        self.gpu_active_flag = gpu_setting["gpu_active"]["0"]
        self.gpu_device = gpu_setting["gpu_device"]["0"]

        if flag_report:
            print("PiSCAT's general parallel GPU flag is set to {}".format(self.gpu_active_flag))
            print("\nThe code is executed on the GPU device {}.".format(self.gpu_device))

    def get_available_devices(self):
        local_device_protos = device_lib.list_local_devices()
        return [
            [x.name, x.memory_limit]
            for x in local_device_protos
            if x.device_type == "GPU" or x.device_type == "CPU"
        ]

    def print_all_available_gpu(self):
        list_gpu_cpu = self.get_available_devices()
        count_ = 0

        for gpu_ in list_gpu_cpu:
            if "GPU" in gpu_[0]:
                count_ = count_ + 1
                print(str(gpu_[0]) + ", memory:" + str(gpu_[1]))

        if count_ == 0:
            print(f"{self.WARNING}\nNo GPU found, TensorFlow will use CPU only!!{self.ENDC}")
